#include <stdio.h>
#include <stdlib.h>

#define TRUE 1
#define FALSE 0

#define MULTIPLE_POS_GAIN FALSE
#define MULTIPLE_HISTORY TRUE

#define MAX_CPUS 8 /* vary for increase, up with 64 */
#define RP_PER_CPU 8 /* per cpu */
#define MAX_SP 16 /* total splitpoints */

#ifdef WINDOWS
#include <windows.h>
#define NOME_WINDOWS
/* typedef ? */
#define sint8 __int8
#define sint16 __int16
#define sint32 __int32
#define sint64 __int64
#define uint8 unsigned __int8
#define uint16 unsigned __int16
#define uint32 unsigned __int32
#define uint64 unsigned __int64
#define INLINE _inline
#define atoll _atoi64
#define TYPE_64_BIT "%I64d"
#if 1
#define MEMALIGN(a, b, c) a = _aligned_malloc (c, b)
#define ALIGNED_FREE(x) _aligned_free (x)
#else
#define MEMALIGN(a, b, c) a = malloc (c)
#define ALIGNED_FREE(x) free (x)
#endif
#define __builtin_prefetch /* */
#ifdef WINDOWS_X64 /* 64 bits */
#include "win64bits.h"
#else
#include "win32bits.h"
#endif
#define MUTEX_TYPE CRITICAL_SECTION
#define COND_TYPE HANDLE
#define LOCK(x) EnterCriticalSection (x)
#define UNLOCK(x) LeaveCriticalSection (x)
#define WAIT_CON_LOCK(x, y) WaitForSingleObject (x, INFINITE)
#define SIGNAL_CON_LOCK(x, y) SetEvent (x)
#define LOCK_INIT(x) InitializeCriticalSection (x)
#define COND_INIT(x, y) (x) = CreateEvent (0, FALSE, FALSE, 0)
#define PTHREAD_CREATE(N, b, thr, d) \
  (*(N)) = CreateThread (NULL, 0, thr, (LPVOID) (d), 0, NULL);
HANDLE PTHREAD[MAX_CPUS], PTHREAD_IO;

#define IVAN_THREAD(A) DWORD WINAPI ivan_thread (LPVOID A)
#define IO_THREAD(A) DWORD WINAPI io_thread (LPVOID A)
#define VOID_STAR_TYPE DWORD
#else /* LINUX */
#define sint8 signed char
#define sint16 signed short int
#define sint32 int
#define sint64 long long int
#define uint8 unsigned char
#define uint16 unsigned short int
#define uint32 unsigned int
#define uint64 unsigned long long int
#define INLINE inline
#define TYPE_64_BIT "%lld"
#define MEMALIGN(a, b, c) posix_memalign ((void*) &(a), b, c)
#define ALIGNED_FREE(x) free (x)
#include "bits.h"
#include <pthread.h>
#define MUTEX_TYPE pthread_mutex_t
#define COND_TYPE pthread_cond_t
#define LOCK(x) pthread_mutex_lock (x)
#define UNLOCK(x) pthread_mutex_unlock (x)
#define WAIT(x, y) pthread_cond_wait (x, y)
#define SIGNAL(x) pthread_cond_signal (x)
#define WAIT_CON_LOCK(x, y) { LOCK (&(y)); WAIT (&(x), &(y)); UNLOCK (&(y)); }
#define SIGNAL_CON_LOCK(x, y) { LOCK (&(y)); SIGNAL (&(x)); UNLOCK (&(y)); }
#define LOCK_INIT(x) pthread_mutex_init ((x), NULL)
#define LOCK_DESTROY pthread_mutex_destroy /* new */
#define COND_INIT(x, y) \
 { pthread_cond_init (&(x), NULL); pthread_mutex_init (&(y), NULL); }
#define PTHREAD_CREATE(N, b, thr, d) pthread_create (N, NULL, thr, (void*) (d))
pthread_t PTHREAD[MAX_CPUS], PTHREAD_IO;
#define IVAN_THREAD(A) void* ivan_thread (void* A)
#define IO_THREAD(A) void* io_thread (void* A)
#define VOID_STAR_TYPE void*
#endif

MUTEX_TYPE PTHREAD_COND_MUTEX[MAX_CPUS], WAKEUP_LOCK_IO[1], SMP_IVAN[1];;
COND_TYPE PTHREAD_COND_WAIT[MAX_CPUS], WAKEUP_IO[1];
#define boolean uint8
boolean volatile DIE[MAX_CPUS];
boolean volatile SMP_HAS_AKTIV;
boolean volatile IVAN_ALL_HALT;
boolean volatile IO_AWAKE;
boolean volatile INFINITE_LOOP;
boolean volatile SEARCH_IS_DONE;
boolean volatile STALL_MODE;
boolean volatile UCI_NEW_GAME;
int SINCE_NEW_GAME;
int EASY_FACTOR, EASY_FACTOR_PONDER, BATTLE_FACTOR, ORDINARY_FACTOR,
  ABSOLUTE_PERCENT, DESIRED_MILLIS, BOOK_EXIT_MOVES;
boolean EXTEND_IN_CHECK;

/****************************************************************/

typedef struct
{
  uint64 volatile PAWN_HASH;
  uint8 wPfile_count, bPfile_count, OpenFileCount;
  boolean locked;
  uint32 wKdanger, bKdanger;
  uint8 wPlight, wPdark, bPlight, bPdark, wPassedFiles, bPassedFiles,
    wDrawWeight, bDrawWeight;
  uint32 SCORE;
} typePawnEval; /* 32 bytes */
typePawnEval *PawnHash;

#define CHECK_HALT() { if (POSITION->stop) { RETURN (0); } }

#define FLAG_LOWER 1
#define FLAG_UPPER 2
#define FLAG_CUT 4
#define FLAG_ALL 8
#define FLAG_EXACT 16
#define FLAG_MOVE_LESS 32
#define UPPER_FROM_PV 64
#define LOWER_FROM_PV 128

#define IsCUT(tr) ((tr->flags) & FLAG_CUT)
#define IsALL(tr) ((tr->flags) & FLAG_ALL)
#define IsExact(tr) ((tr)->flags & FLAG_EXACT)

typedef struct
{
  /* 00-03 */ uint32 hash;
  /* 04-07 */ uint8 flags, age, DepthUpper, DepthLower;
  /* 08-0b */ sint16 ValueLower, ValueUpper; /* overlap hash in lockless */
  /* 0c-0d */ uint16 move;
  /* 0e-0f */ uint8 rev, moveless;
} typeHash;

typeHash *HashTable;
uint64 HashMask, AGE;

#define HYPER_HASH /* for the novel! */
#ifndef HYPER_HASH
#define HASH_POINTER(x) (HashTable + ((x) & HashMask))
#define HASH_POINTER0(x) (HashTable + ((x) & (3 + HashMask)))
#else
uint8 HASH_WHICH[16];
typeHash* TABLE_ARR[3];
uint64 MASK_ARR[3]; /* 64 bytes, but yet for the align? */
#define HASH_POINTER(x) \
  (TABLE_ARR[HASH_WHICH[(x) >> 60]] + ((x) & MASK_ARR[HASH_WHICH[(x) >> 60]]))
#define HASH_POINTER0(x) \
  (TABLE_ARR[HASH_WHICH[(x) >> 60]] + ((x) & (3 + MASK_ARR[HASH_WHICH[(x) >> 60]])))
#endif

#define NEW_HASH
#ifdef NEW_HASH
#define VAS(v) VALUE_ADJUST_STORE((v), (HEIGHT(POSITION)))
#define HashLowerALL(POS, mv, dp, v) \
  HashLowerALLNew (POS->DYN->HASH, mv, dp, VAS (v))
#define HashUpperCUT(POS, dp, v) \
  HashUpperCUTNew (POS->DYN->HASH, dp, VAS (v))
#define HashLower(Z, mv, dp, v) HashLowerNew (Z, mv, dp, VAS (v), 0)
#define HashUpper(Z, dp, v) HashUpperNew (Z, dp, VAS (v), 0)
#define HashLowerPV(Z, mv, dp, v) HashLowerNew (Z, mv, dp, VAS (v), LOWER_FROM_PV)
#define HashUpperPV(Z, dp, v) HashUpperNew (Z, dp, VAS (v), UPPER_FROM_PV)
#define HashExact(POS, mv, dp, v, FL) \
  HashExactNew (POS->DYN->HASH, mv, dp, VAS (v), FL)
#endif

#if 0
#define MATEING_SCORE /* Fact: slows by 1-2%: for the why? */
#endif
#ifdef MATEING_SCORE
#if 1 /* asm code */ /* still the snail? */
static INLINE sint32 VALUE_ADJUST_STORE(sint32 v, uint32 ht)
{
  asm volatile ("xor %%r8d,%%r8d\n"
		"cmpl $29000,%0\n"
		"cmovg %1,%%r8d\n"
		"xor %%r9d,%%r9d\n"
		"cmpl $-29000,%0\n"
		"cmovl %1,%%r9d\n"
		"add %%r8d,%0\n"
		"sub %%r9d,%0\n" :
		"=r" (v) :
		"r" (ht), "0" (v) :
		"%r8", "%r9");
 return v;
}
#define HashUpperBound(tr) \
  VALUE_ADJUST_STORE (tr->ValueUpper, -HEIGHT(POSITION))
#define HashLowerBound(tr) \
  VALUE_ADJUST_STORE (tr->ValueLower, -HEIGHT(POSITION))
#else /* no asm code, slower */
#define VALUE_ADJUST_STORE(val, ht) \
  ((val) + (((val) > 29000) ? (ht) : 0) - (((val) < -29000) ? (ht) : 0))
#define HashUpperBound(tr) \
  (tr->ValueUpper - ((tr->ValueUpper > 29000) ? (HEIGHT(POSITION)) : 0)	\
                  + ((tr->ValueUpper < -29000) ? (HEIGHT(POSITION)) : 0))
#define HashLowerBound(tr) \
  (tr->ValueLower - ((tr->ValueLower > 29000) ? (HEIGHT(POSITION)) : 0)	\
                  + ((tr->ValueLower < -29000) ? (HEIGHT(POSITION)) : 0))
#endif
#else
#define VALUE_ADJUST_STORE(Value, height) Value
#define HashUpperBound(tr) (tr->ValueUpper)
#define HashLowerBound(tr) (tr->ValueLower)
#endif

#define FILE(s) ((s) & 7)
#define RANK(s) ((s) >> 3)
#define FROM(s) (((s) >> 6) & 077)
#define TO(s) ((s) & 077)

/* YUSUF */

boolean TITANIC_MODE;
int volatile CPUS_SIZE;
uint64 NODE_CHECK;

#if 0
#define LOCKLESS_HASHING /* functional Yet: poored */
#endif
#ifdef LOCKLESS_HASHING  /* Fact: 2-3% slowed, for the why? */
#define TRANS_DECLARE() typeHash trans[1]
#define YUSUF_DECLARE() typeHash *Trans
#define TRANS_PV_DECLARE() typePVHash trans_pv[1]
#define YUSUF_PV_DECLARE() typePVHash *Trans_pv
#if 0 /* on/off */
#define READ_WRITE_BARRIER() asm volatile ("" ::: "memory")
#define HYATT_HASH(T, t) \
  { READ_WRITE_BARRIER(); \
    ((uint64*) (t))[0] = ((uint64*) (T))[0] ^ ((uint64*) (T))[1]; \
    ((uint64*) (t))[1] = ((uint64*) (T))[1]; READ_WRITE_BARRIER(); }
#define HYATT_HASH_WRITE(t, T) \
  { READ_WRITE_BARRIER(); \
    ((uint64*) (T))[0] = ((uint64*) (t))[0] ^ ((uint64*) (t))[1]; \
    ((uint64*) (T))[1] = ((uint64*) (t))[1]; READ_WRITE_BARRIER(); }
#else /* dummy */ /* Fact: 2-3% snailed in the already with the this */
#define HYATT_HASH(T, t) { memcpy (t, T, 16); }
#define HYATT_HASH_WRITE(T, t) { memcpy (t, T, 16); }
#endif
#define UPDATE_AGE() /* ? */ /* Need for the think */
#else /* Personal Opinion: in the forget for the LOCKLESS_HASHING */
#define Trans trans
#define Trans_pv trans_pv
#define TRANS_DECLARE() typeHash *trans
#define YUSUF_DECLARE() /* */
#define TRANS_PV_DECLARE() typePVHash *trans_pv
#define YUSUF_PV_DECLARE() /* */
#define HYATT_HASH(x, y) /* */
#define HYATT_HASH_WRITE(x, y) /* */
#if 1
#define UPDATE_AGE() trans->age = AGE; /* purposeful? */
#else
#define UPDATE_AGE() /* */
#endif
#endif

#define EVAL(m) Eval (POSITION, -0x7fff0000, 0x7fff0000, m)
#define EvalLazy(B, A, p, m) Eval (POSITION, (B) - (p), (A) + (p), m)

#define WHITE_IN_CHECK (POSITION->DYN->bAtt & wBitboardK)
#define BLACK_IN_CHECK (POSITION->DYN->wAtt & bBitboardK)
#define POS1 (POS0 + 1)
#define MOVE_IS_CHECK_WHITE (POS1->wKcheck)
#define MOVE_IS_CHECK_BLACK (POS1->bKcheck)

#ifdef MULTIPLE_POS_GAIN
sint16 MAX_POSITIONAL_GAIN[MAX_CPUS][0x10][010000];
#define MAX_POS_GAIN(pez, mos) MAX_POSITIONAL_GAIN[POSITION->cpu][pez][mos]
#else
sint16 MAX_POSITIONAL_GAIN[0x10][010000]; /* multicore ? */
#define MAX_POS_GAIN(pez, mos) MAX_POSITIONAL_GAIN[pez][mos]
#endif

boolean BOARD_IS_OK, NEW_GAME;

#define CheckRepetition \
  CHECK_HALT(); \
  if (POSITION->DYN->reversible >= 100) RETURN(0); \
  for (i = 4; i <= POSITION->DYN->reversible && \
	      i <= POSITION->StackHeight; i += 2) \
    if (POSITION->STACK[POSITION->StackHeight - i] == POSITION->DYN->HASH) RETURN(0);

typedef struct
{
  uint64 hash;
  sint32 Value;
  uint16 move;
  uint8 depth, age;
} typePVHash;
typePVHash *PVHashTable; /* multicore ? */
uint64 PVHashMask;

int PawnHashSize; /* uci option ahora */

typedef struct
{
  uint32 move;
} typeMoveList;

typedef struct
{
  uint32 move;
  sint32 value;
  uint64 nodes;
} typeRootMoveList;

typedef struct
{
  /* 0x00-0x0f */ uint64 HASH, PAWN_HASH;
  /* 0x10-0x1b */ uint32 material, STATIC, _7; /* STATIC es sint32 ? */
  /* 0x1c-0x1f */  uint8 oo, reversible, ep, cp;
  /* 0x20-0x3f */ uint64 wAtt, bAtt, wXray, bXray;
  /* 0x40-0x47 */ sint32 Value, PositionalValue;
  /* 0x48-0x51 */ uint16 _5, _6, killer1, killer2, move;
  /* 0x52-0x57 */  uint8 _0, _3, exact, lazy, SAVED_FLAGS, flags;
  /* 0x58-0x7f */ uint64 wKcheck, bKcheck, _1, _2, _8;
} typeDYNAMIC;

typedef struct
{
  int phase, mask, bc;
  uint32 trans_move, move, exclude;
  uint64 TARGET;
  typeMoveList LIST[256];
  uint32 BAD_CAPS[64];
} typeNEXT;

#define HEIGHT(x) ((x)->height)

typedef struct
{
  int alpha;
  int beta;
  int depth;
  int node_type;
  int value;
  uint32 move;
  uint32 good_move;
  uint32 childs;
  typeNEXT* MOVE_PICK;
  boolean tot;
  boolean aktiv;
  MUTEX_TYPE splock[1];
}
SPLITPUNKT;

SPLITPUNKT ROOT_SP[MAX_SP];
volatile uint64 SMP_FREE;
#define NODE_TYPE_PV 1
#define NODE_TYPE_ALL 2
#define NODE_TYPE_CUT 3
#include <setjmp.h>

#define MAX_STACK 256
struct TP
{
  uint8 sq[64];
  uint64 bitboard[16];
  uint64 OccupiedBW, OccupiedL90, OccupiedL45, OccupiedR45;
  uint8 XRAYw[64], XRAYb[64];
  uint8 wtm, wKsq, bKsq, height;
  /* Above bytes are copied, exactly 356 of them */
  /* 64 + 128 + 32 + 128 + 4 = 356 */
  typeDYNAMIC *DYN, *DYN_ROOT;
  uint64 STACK[MAX_STACK], StackHeight;
  uint64 nodes, tbhits;
  uint8 cpu, rp;
  boolean stop, used;
  MUTEX_TYPE padlock[1];
  int child_count;
  struct TP *parent, *children[MAX_CPUS];
  SPLITPUNKT *SplitPoint;
};

typedef struct TP typePOS;
typePOS ROOT_POSITION[MAX_CPUS][RP_PER_CPU];
typePOS ROOT_POSIZIONE0[1], NULL_PARENT[1];

/***************************** ZOG MP *******************************/

typedef struct {uint16 move; sint16 score;} type_ZOG_ML;
struct ZN
{
  typePOS POS[1]; /* STACK has 2048 bytes */
  MUTEX_TYPE lock[1];
  int alpha, beta, depth, orig_alpha, expected_value, best_value;
  struct ZN *PARENT;
  int USERS, AKTIV_MOVES_COUNT;
  int zm, phase;
  int ZOG_MOVE_COUNT, TOTAL_MOVE_COUNT, extension;
  boolean cpus_used[MAX_CPUS];
  struct ZN *children[MAX_CPUS];
  uint32 MOVE_FROM_PARENT, best_move;
  boolean is_busy, dead, EXPANDED, join_able, is_registered, is_book_keep;
  uint64 hash, when;
  int num, previous_result;
  type_ZOG_ML MOVE_LIST[256]; /* 1024 bytes */
}; /* volatile? */

boolean ZOG_AKTIV;
uint64 ZOG_SIZE, ZOG_HEFT; /* int ? */
boolean volatile ZOG_AVAILABLE;
boolean volatile ZOG_AVAILABLE_OTHER;
boolean volatile ZOG_ROOT_DONE;
typedef struct ZN type_zog_node;
#define MIN_ZOG_DEPTH 10

/*********************************/ /*********************************/

#define MAXIMUM_PLY 2048
/* SMP rotogravure */
#define WhiteOO (POSITION->DYN->oo & 0x1)
#define WhiteOOO (POSITION->DYN->oo & 0x2)
#define BlackOO (POSITION->DYN->oo & 0x4)
#define BlackOOO (POSITION->DYN->oo & 0x8)

typedef struct { sint16 Value; uint8 token, flags; } typeMATERIAL;
typeMATERIAL MATERIAL[419904]; /* SMP read */

typedef enum
{ wEnumOcc, wEnumP, wEnumN, wEnumK, wEnumBL, wEnumBD, wEnumR, wEnumQ,
  bEnumOcc, bEnumP, bEnumN, bEnumK, bEnumBL, bEnumBD, bEnumR, bEnumQ
} EnumPieces;
typedef enum
{ TRANS, CAPTURE_GEN, CAPTURE_MOVES,
    KILLER1, KILLER2, ORDINARY_MOVES, BAD_CAPS,
  TRANS2, CAPTURE_PGEN2, CAPTURE_MOVES2, QUIET_CHECKS, EVADE_PHASE, TRANS3,
  CAPTURE_GEN3, CAPTURE_MOVES3, QUIET_CHECKS3, POSITIONAL_GAIN_PHASE, FASE_0
} EnumPhases;
typedef enum
{ A1, B1, C1, D1, E1, F1, G1, H1, A2, B2, C2, D2, E2, F2, G2, H2,
  A3, B3, C3, D3, E3, F3, G3, H3, A4, B4, C4, D4, E4, F4, G4, H4,
  A5, B5, C5, D5, E5, F5, G5, H5, A6, B6, C6, D6, E6, F6, G6, H6,
  A7, B7, C7, D7, E7, F7, G7, H7, A8, B8, C8, D8, E8, F8, G8, H8
} EnumSquares;
typedef enum { R1, R2, R3, R4, R5, R6, R7, R8 } EnumRanks;
typedef enum { FA, FB, FC, FD, FE, FF, FG, FH } EnumFiles;

sint32 PieceSquareValue[0x10][0100]; /* SMP read */

#ifdef MULTIPLE_HISTORY
uint16 HISTORY[MAX_CPUS][0x10][0100]; /* SMP cashline ? */
#define HISTORY_VALUE(P, M) HISTORY[P->cpu][P->sq[FROM (M)]][TO (M)]
#else
uint16 HISTORY[0x10][0100];
#define HISTORY_VALUE(P, M) HISTORY[P->sq[FROM (M)]][TO (M)]
#endif

char STRING1[MAX_CPUS][64], STRING2[MAX_CPUS][64],
  STRING3[MAX_CPUS][64], STRING4[MAX_CPUS][64]; /* SMP */

#define wBitboardK POSITION->bitboard[wEnumK]
#define wBitboardQ POSITION->bitboard[wEnumQ]
#define wBitboardR POSITION->bitboard[wEnumR]
#define wBitboardBL POSITION->bitboard[wEnumBL]
#define wBitboardBD POSITION->bitboard[wEnumBD]
#define wBitboardB (wBitboardBL|wBitboardBD)
#define wBitboardN POSITION->bitboard[wEnumN]
#define wBitboardP POSITION->bitboard[wEnumP]
#define wBitboardOcc POSITION->bitboard[wEnumOcc]
#define bBitboardK POSITION->bitboard[bEnumK]
#define bBitboardQ POSITION->bitboard[bEnumQ]
#define bBitboardR POSITION->bitboard[bEnumR]
#define bBitboardBL POSITION->bitboard[bEnumBL]
#define bBitboardBD POSITION->bitboard[bEnumBD]
#define bBitboardB (bBitboardBL|bBitboardBD)
#define bBitboardN POSITION->bitboard[bEnumN]
#define bBitboardP POSITION->bitboard[bEnumP]
#define bBitboardOcc POSITION->bitboard[bEnumOcc]

#define ShiftLeft45 LineShift[Direction_h1a8]
#define ShiftRight45 LineShift[Direction_a1h8]
#define ShiftAttack LineShift[Direction_horz]
#define ShiftLeft90 LineShift[Direction_vert]
#define AttLeft45 LineMask[Direction_h1a8]
#define AttRight45 LineMask[Direction_a1h8]
#define AttNormal LineMask[Direction_horz]
#define AttLeft90 LineMask[Direction_vert]

#define Att_h1a8(fr) \
 AttLeft45[fr][(POSITION->OccupiedL45 >> ShiftLeft45[fr]) & 077]
#define Att_a1h8(fr) \
 AttRight45[fr][(POSITION->OccupiedR45 >> ShiftRight45[fr]) & 077]
#define AttRank(fr) \
 AttNormal[fr][(POSITION->OccupiedBW >> ShiftAttack[fr]) & 077]
#define AttFile(fr) \
 AttLeft90[fr][(POSITION->OccupiedL90 >> ShiftLeft90[fr]) & 077]

#define MAX(x, y) (( (x) >= (y)) ? (x) : (y))
#define MIN(x, y) (( (x) <= (y)) ? (x) : (y))
#define ABS(x) (( (x) >= 0) ? (x) : -(x))
#define FileDistance(x, y) (ABS(FILE(x) - FILE(y)))
#define RankDistance(x, y) (ABS(RANK(x) - RANK(y)))

#define AttB(fr) (Att_a1h8(fr) | Att_h1a8(fr))
#define AttR(fr) (AttRank(fr) | AttFile(fr))
#define AttQ(fr) (AttR(fr) | AttB(fr))

#define FlagEP 030000
#define FlagOO 010000
#define FLAG_MASK 070000
#define FlagPromQ 070000
#define FlagPromR 060000
#define FlagPromB 050000
#define FlagPromN 040000
#define MoveIsEP(x) (((x) & FLAG_MASK) == FlagEP)
#define MoveIsProm(x) (((x) & FLAG_MASK) >= FlagPromN)
#define MoveIsOO(x) (((x) & FLAG_MASK) == FlagOO)
#define MoveHistory(x) (((x) & 060000) == 0)

#define Direction_h1a8 0
#define Direction_a1h8 1
#define Direction_horz 2
#define Direction_vert 3
#define BAD_DIRECTION 37
#define VALUE_MATE 30000
#define VALUE_INFINITY 32750
#define MOVE_NONE 0

#ifdef USER_SPLIT
boolean SPLIT_AT_CUT;
int CUT_SPLIT_DEPTH, ALL_SPLIT_DEPTH, PV_SPLIT_DEPTH;
#else
#define SPLIT_AT_CUT TRUE
#define CUT_SPLIT_DEPTH 16
#define ALL_SPLIT_DEPTH 14
#define PV_SPLIT_DEPTH 14
#endif

volatile boolean DO_PONDER;
volatile boolean PONDER_HIT;
volatile boolean DO_INFINITE;
volatile boolean SUPPRESS_INPUT;
volatile boolean STOP; /* de sorvegliare.c */

boolean DO_SEARCH_MOVES, SEND_CURR_MOVE, ROBBO_LOAD;

uint32 SEARCH_MOVES[256];

typedef struct
{ uint32 move; sint32 Value, alpha, beta; uint32 depth, _0; uint64 nodes; } typeMPV;
typeMPV MPV[256];
int MULTI_PV;

#ifndef MINIMAL
int RANDOM_COUNT,RANDOM_BITS, MULTI_CENTI_PAWN_PV;
boolean ALWAYS_ANALYZE, TRY_PV_IN_ANALYSIS, DEBUG_TIME_MANAGEMENT;
#endif
int UCI_White_Bishops_Scale, UCI_White_Pawn_Scale, UCI_White_Knight_Scale;
int UCI_White_Light_Scale, UCI_White_Dark_Scale;
int UCI_White_Rook_Scale, UCI_White_Queen_Scale;
int UCI_Black_Bishops_Scale, UCI_Black_Pawn_Scale, UCI_Black_Knight_Scale;
int UCI_Black_Light_Scale, UCI_Black_Dark_Scale;
int UCI_Black_Rook_Scale, UCI_Black_Queen_Scale;
int UCI_MATERIAL_WEIGHTING, UCI_PAWNS_WEIGHTING, UCI_KING_SAFETY_WEIGHTING;

boolean USE_ALTERNATIVE_TIME, ALLOW_INSTANT_MOVE, SEARCH_ROBBO_BASES;
boolean DO_OUTPUT, UCI_PONDER, VERIFY_NULL, SEND_HASH, DO_HASH_FULL;
int BUFFER_TIME, OUTPUT_DELAY;
char ROBBO_TRIPLE_DIR[1024], ROBBO_TOTAL_DIR[1024];

#include "arrays.h"
#include "functions.h"
#include "common.h"

#ifdef DEBUG
#define UTILITIES
#define VALIDATE(a, x, y) validate (a, x, y)
#else
#define VALIDATE(a, x, y) /* */
#endif

#define ROBBO_PICCOLO TRUE
#define ROBBO_SCACCO TRUE

#define IS_EXACT(x) (x) /* exact from RobboTriple */

#if 0
#define TRACE_COMPILE
#endif
#ifdef TRACE_COMPILE
boolean TRACE_QSEARCH, TRACE_HASH, TRACE_EVAL, TRACE_LOW_DEPTH, TRACE_CUT,
  TRACE_ALL, TRACE_EXCLUDE, TRACE_PV, TRACE_PVQSEARCH, TRACE_MAKE_UNMAKE;
#define TRACE(x, y) if (x) {y;}
#define HASH_READ(t) \
  printf ("READ [%x] %llx %x %x %s U:%d|%d L:%d|%d\n", \
	  k + i, POSITION->DYN->HASH, t->flags, t->age, Notate(t->move, STRING1[POSITION->cpu]), \
	  t->DepthUpper, t->ValueUpper, t->DepthLower, t->ValueLower);
#define HASH_READZ(t, zob) \
  printf ("READ [%x] %llx %x %x %s U:%d|%d L:%d|%d\n", \
	  k + i, zob, t->flags, t->age, Notate(t->move, STRING1[POSITION->cpu]), \
	  t->DepthUpper, t->ValueUpper, t->DepthLower, t->ValueLower);
#define HASH_WRITE(s, t) \
  printf ("HASH %s [%x] %llx %x %x %s U:%d|%d L:%d|%d\n", \
	  s, k + i, Z, t->flags, t->age, Notate(t->move, STRING1[POSITION->cpu]), \
	  t->DepthUpper, t->ValueUpper, t->DepthLower, t->ValueLower);
#define HASH_OVERWRITE(s, t) \
  printf ("HASHo %s [%x] %llx %x %x %s U:%d|%d L:%d|%d\n", \
	  s, k + w, Z, t->flags, t->age, Notate(t->move, STRING1[POSITION->cpu]), \
	  t->DepthUpper, t->ValueUpper, t->DepthLower, t->ValueLower);
#define HASH_ZERO(t) printf ("HASHz [%x]\n", k + j);
#else
#define TRACE(x, y) /* */
#endif

#if 0
#define ROBBO_TRIPLE_CONDITION(POSITION) (POPCNT (POSITION->OccupiedBW) <= 5)
#define ROBBO_TOTAL_CONDITION(POSITION) (POPCNT (POSITION->OccupiedBW) <= 6)
#else
#define ROBBO_TRIPLE_CONDITION(POSITION) \
  (POPCNT (POSITION->OccupiedBW ^ (wBitboardP & (bBitboardP >> 8))) <= 5)
#define ROBBO_TOTAL_CONDITION(POSITION) \
  (POPCNT (POSITION->OccupiedBW ^ (wBitboardP & (bBitboardP >> 8))) <= 6)
#endif

FILE* LOGFILE;
char *INPUT_BUFFER, *input_ptr;

