
#include "RobboLito.h"
#include "RobboTotalBase.h"

#include <sys/time.h>
uint64
GetClock ()
{
  uint64 x;
  struct timeval tv;
  gettimeofday (&tv, NULL);
  x = tv.tv_sec;
  x *= 1000000;
  x += tv.tv_usec;
  return x;
}

static uint64 RAGGIO_NO[64], RAGGIO_SE[64], RAGGIO_SO[64], RAGGIO_NE[64];
static uint64 RAGGIO_N[64], RAGGIO_S[64], RAGGIO_O[64], RAGGIO_E[64];

#define NUOVO_F1 TRUE
#define NUOVO_F2 TRUE

void DISCO_FREAD1 (uint8 *A, uint64 s, uint64 pro, FILE *F, uint64 fpro)
{
#if !NUOVO_F1
  fread (A, s, pro, F);
#else
  fread (A, s, fpro, F);
  dehuffman1 (A, pro);
#endif
}

void DISCO_FREAD2 (uint8 *A, uint64 s, uint64 pro, FILE *F, uint64 fpro)
{
#if !NUOVO_F2
  fread (A, s, pro, F);
#else
  fread (A, s, fpro, F);
  if (pro != fpro)
    dehuffman2 (A, pro);
#endif
}

int DISCO_FWRITE1 (uint8 *A, uint64 s, uint64 pro, FILE *F)
{
#if !NUOVO_F1
  return fwrite (A, s, pro, F);
#else
  uint8 *B;
  B = malloc (pro);
  int w = huffman1 (A, B, pro);
  fwrite (B, s, w, F);
  free (B);
  return w;
#endif
}

int DISCO_FWRITE2 (uint8 *A, uint64 s, uint64 pro, FILE *F)
{
#if !NUOVO_F2
  return fwrite (A, s, pro, F);
#else
  uint8 *B;
  B = malloc (2 * pro);
  int w = huffman2 (A, B, pro);
  fwrite (B, s, w, F);
  free (B);
  return w;
#endif
}

static void
iniz_raggio ()
{
  int sq, tr, co, TR, CO;
  for (sq = A1; sq <= H8; sq++)
    {
      TR = RANK (sq);
      CO = FILE (sq);
      RAGGIO_NE[sq] = 0;
      for (tr = TR, co = CO; tr <= R8 && co <= FH; tr++, co++)
	RAGGIO_NE[sq] |= SqSet[8 * tr + co];
      RAGGIO_NO[sq] = 0;
      for (tr = TR, co = CO; tr <= R8 && co >= FA; tr++, co--)
	RAGGIO_NO[sq] |= SqSet[8 * tr + co];
      RAGGIO_SO[sq] = 0;
      for (tr = TR, co = CO; tr >= R1 && co >= FA; tr--, co--)
	RAGGIO_SO[sq] |= SqSet[8 * tr + co];
      RAGGIO_SE[sq] = 0;
      for (tr = TR, co = CO; tr >= R1 && co <= FH; tr--, co++)
	RAGGIO_SE[sq] |= SqSet[8 * tr + co];
      RAGGIO_N[sq] = 0;
      for (tr = TR; tr <= R8; tr++)
	RAGGIO_N[sq] |= SqSet[8 * tr + CO];
      RAGGIO_S[sq] = 0;
      for (tr = TR; tr >= R1; tr--)
	RAGGIO_S[sq] |= SqSet[8 * tr + CO];
      RAGGIO_O[sq] = 0;
      for (co = CO; co >= FA; co--)
	RAGGIO_O[sq] |= SqSet[8 * TR + co];
      RAGGIO_E[sq] = 0;
      for (co = CO; co <= FH; co++)
	RAGGIO_E[sq] |= SqSet[8 * TR + co];
      RAGGIO_NE[sq] &= ~SqSet[sq];
      RAGGIO_SO[sq] &= ~SqSet[sq];
      RAGGIO_N[sq] &= ~SqSet[sq];
      RAGGIO_S[sq] &= ~SqSet[sq];
      RAGGIO_SE[sq] &= ~SqSet[sq];
      RAGGIO_NO[sq] &= ~SqSet[sq];
      RAGGIO_E[sq] &= ~SqSet[sq];
      RAGGIO_O[sq] &= ~SqSet[sq];
    }
}

void
FORZA_POS (RobboTotalBase * TotalBase, uint64 PositionCount, tipo_fPOS * fpos)
{
  int i, rf, m, wK, bK, v, x;
  uint64 n = PositionCount, q, r, sq;
  boolean PEDONE = FALSE;
  int PI[4] = { 0, 0, 0, 0 }, Squ[4] = { 0, 0, 0, 0};

  for (i = 3; i >= 0; i--)
    {
      if (TotalBase->p[i] == wEnumP || TotalBase->p[i] == bEnumP
	  || TotalBase->p[i] == BlockedPawn)
	PEDONE = TRUE;
      DIV (q, n, TotalBase->m[i]);
      r = n - q * TotalBase->m[i];
      sq = r;
      n = q;
#if 1
      if (TotalBase->p[i] && i < 1 && (TotalBase->p[i]) == (TotalBase->p[i + 3]))
	{
	  if (TotalBase->p[i] == wEnumP || TotalBase->p[i] == bEnumP)
	    m = quad_quat_pedone[sq];
	  else if (TotalBase->p[i] == BlockedPawn)
	    m = quad_quat_blocked[sq];
	  else
	    m = quad_quat_normale[sq];
	  Squ[i] = m & 63;
	  Squ[i + 1] = (m >> 6) & 63;
	  Squ[i + 2] = (m >> 12) & 63;
	  Squ[i + 3] = m >> 18;
	}
#endif
      else if (TotalBase->p[i] && i < 2 && (TotalBase->p[i]) == (TotalBase->p[i + 2]))
	{
	  if (TotalBase->p[i] == wEnumP || TotalBase->p[i] == bEnumP)
	    m = quad_tre_pedone[sq];
	  else if (TotalBase->p[i] == BlockedPawn)
	    m = quad_tre_blocked[sq];
	  else
	    m = quad_tre_normale[sq];
	  Squ[i] = m & 63;
	  Squ[i + 1] = (m >> 6) & 63;
	  Squ[i + 2] = m >> 12;
	}
      else if (TotalBase->p[i] && i < 3 && (TotalBase->p[i]) == (TotalBase->p[i + 1])
	       && TotalBase->m[i] != 1)
	{
	  if (TotalBase->p[i] == wEnumP || TotalBase->p[i] == bEnumP)
	    m = Sq2Pawn[sq];
	  else if (TotalBase->p[i] == BlockedPawn)
	    m = Sq2Blocked[sq];
	  else
	    m = Sq2Normal[sq];
	  Squ[i] = m & 63;
	  Squ[i + 1] = m >> 6;
	}
      else if (TotalBase->m[i] != 1)
	{
	  if (TotalBase->m[i] == 48  || TotalBase->m[i] == 40)
	    sq += 8;
	  Squ[i] = sq;
	}
    }
  if (PEDONE)
    rf = inverso_pedone_re[n];
  else
    rf = inverso_normale_re[n];
  for (i = 0; i < 4; i++)
    PI[i] = col_scamb[TotalBase->p[i]];
  wK = rf & 63;
  bK = rf >> 6;
  fpos->PEDONE = PEDONE;
  if (PEDONE)
    {
      int w = 2 * (FILE (bK) >= FE) + 1;
      for (i = 0; i < 4; i++)
	{
	  LANCpez (w, Squ[i]);
	  if (PI[i] == BlockedPawn)
	    Squ[i] -= 8;
	}
      LANCpez (w, wK);
      LANCpez (w, bK);
      rf = RE_pedone[64 * wK + bK];
    }
  else
    {
      uint8 w = oLANC[wK][bK];
      for (i = 0; i < 4; i++)
	LANCnor (w, Squ[i]);
      LANCnor (w, bK);
      LANCnor (w, wK);
      rf = RE_normale[64 * wK + bK];
    }
  fpos->re_fetta = rf;
  fpos->wK = wK;
  fpos->bK = bK;
  fpos->Occupied = SqSet[wK] | SqSet[bK];
  v = TotalBaseTable[PI[0]][PI[1]][PI[2]][PI[3]];
  for (i = 0; i < 4; i++)
    {
      x = (v >> (16 + 2 * i)) & 3;
      fpos->sq[i] = Squ[x];
      fpos->pi[i] = PI[x];
      if (PI[x])
	fpos->Occupied |= SqSet[Squ[x]];
      if (PI[x] == BlockedPawn)
	fpos->Occupied |= SqSet[Squ[x] + 8];
    }
  fpos->PawnOnFourth = 0;
#if 0
  for (i = 0; i < 4; i++)
    if (fpos->pi[i] == bEnumP && RANK (fpos->sq[i]) == R4)
      fpos->PawnOnFourth |= SqSet[fpos->sq[i]];
#else
  for (i = 0; i < 4; i++)
    if (fpos->pi[i] == bEnumP && RANK (fpos->sq[i]) == R4)
      fpos->PawnOnFourth |= SqSet[fpos->sq[i]];
  for (i = 0; i < 4; i++)
    if ((fpos->pi[i] == BlockedPawn && RANK (fpos->sq[i]) == R3))
      fpos->PawnOnFourth |= SqSet[fpos->sq[i] + 8];
#endif
  fpos->m[3] = 1;

  for (i = 0; i < 4; i++)
    fpos->rip[i] = 0;
  if (fpos->pi[0] && fpos->pi[0] == fpos->pi[3])
    {
      fpos->rip[0] = fpos->rip[1] = fpos->rip[2] = fpos->rip[3] = 4;
      fpos->rip_ai[0] = 1;
      fpos->rip_ai[1] = 2;
      fpos->rip_ai[2] = 3;
      fpos->rip_ai[3] = 4;
    }
  else if (fpos->pi[0] && fpos->pi[0] == fpos->pi[2])
    {
      fpos->rip[0] = fpos->rip[1] = fpos->rip[2] = 3;
      fpos->rip_ai[0] = 1;
      fpos->rip_ai[1] = 2;
      fpos->rip_ai[2] = 3;
    }
  else if (fpos->pi[1] && fpos->pi[1] == fpos->pi[3])
    {
      fpos->rip[1] = fpos->rip[2] = fpos->rip[3] = 3;
      fpos->rip_ai[1] = 1;
      fpos->rip_ai[2] = 2;
      fpos->rip_ai[3] = 3;
    }
  else
    {
      if (fpos->pi[0] && fpos->pi[0] == fpos->pi[1])
	{
	  fpos->rip[0] = fpos->rip[1] = 2;
	  fpos->rip_ai[0] = 1;
	  fpos->rip_ai[1] = 2;
	}
      if (fpos->pi[1] && fpos->pi[1] == fpos->pi[2])
	{
	  fpos->rip[1] = fpos->rip[2] = 2;
	  fpos->rip_ai[1] = 1;
	  fpos->rip_ai[2] = 2;
	}
      if (fpos->pi[2] && fpos->pi[2] == fpos->pi[3])
	{
	  fpos->rip[2] = fpos->rip[3] = 2;
	  fpos->rip_ai[2] = 1;
	  fpos->rip_ai[3] = 2;
	}
    }

  for (i = 2; i >= 0; i--)
    fpos->m[i] =
      fpos->m[i + 1] * ((TABLE_TOTAL_BASES + (v & 65535))->m[i + 1]);
  fpos->re_fetta_molt = fpos->m[0] * (TABLE_TOTAL_BASES + (v & 65535))->m[0];
  for (i = 0; i < 4; i++)
    Squ[i] = fpos->sq[i]
      - 8 * (fpos->pi[i] == wEnumP || fpos->pi[i] == bEnumP)
      - 8 * (fpos->pi[i] == BlockedPawn);
  if (fpos->pi[0] && fpos->pi[0] == fpos->pi[3])
    {
      Squ[0] = STESSO4[Squ[0]][Squ[1]][Squ[2]][Squ[3]];
      Squ[1] = Squ[2] = Squ[3] = 0;
    }
  else if (fpos->pi[0] && fpos->pi[0] == fpos->pi[2])
    {
      Squ[0] = STESSO3[Squ[0]][Squ[1]][Squ[2]];
      Squ[1] = Squ[2] = 0;
    }
  else if (fpos->pi[1] && fpos->pi[1] == fpos->pi[3])
    {
      Squ[1] = STESSO3[Squ[1]][Squ[2]][Squ[3]];
      Squ[2] = Squ[3] = 0;
    }
  else
    {
      if (fpos->pi[0] && fpos->pi[0] == fpos->pi[1])
	{
	  Squ[0] = STESSO2[Squ[0]][Squ[1]];
	  Squ[1] = 0;
	}
      if (fpos->pi[1] && fpos->pi[1] == fpos->pi[2])
	{
	  Squ[1] = STESSO2[Squ[1]][Squ[2]];
	  Squ[2] = 0;
	}
      if (fpos->pi[2] && fpos->pi[2] == fpos->pi[3])
	{
	  Squ[2] = STESSO2[Squ[2]][Squ[3]];
	  Squ[3] = 0;
	}
    }
  fpos->ind = 0;
  for (i = 0; i < 4; i++)
    if (PI[i])
      fpos->ind += fpos->m[i] * Squ[i];
  fpos->Double = -1;
  if ((!PEDONE) && RANK (wK) == FILE (wK) && RANK (bK) == FILE (bK))
    {
      for (i = 0; i < 4; i++)
	Squ[i] = lanc_norm[1][fpos->sq[i]];
      if (fpos->pi[0] && fpos->pi[0] == fpos->pi[3])
	{
	  Squ[0] = STESSO4[Squ[0]][Squ[1]][Squ[2]][Squ[3]];
	  Squ[1] = Squ[2] = Squ[3] = 0;
	}
      else if (fpos->pi[0] && fpos->pi[0] == fpos->pi[2])
	{
	  Squ[0] = STESSO3[Squ[0]][Squ[1]][Squ[2]];
	  Squ[1] = Squ[2] = 0;
	}
      else if (fpos->pi[1] && fpos->pi[1] == fpos->pi[3])
	{
	  Squ[1] = STESSO3[Squ[1]][Squ[2]][Squ[3]];
	  Squ[2] = Squ[3] = 0;
	}
      else
	{
	  if (fpos->pi[0] && fpos->pi[0] == fpos->pi[1])
	    {
	      Squ[0] = STESSO2[Squ[0]][Squ[1]];
	      Squ[1] = 0;
	    }
	  if (fpos->pi[1] && fpos->pi[1] == fpos->pi[2])
	    {
	      Squ[1] = STESSO2[Squ[1]][Squ[2]];
	      Squ[2] = 0;
	    }
	  if (fpos->pi[2] && fpos->pi[2] == fpos->pi[3])
	    {
	      Squ[2] = STESSO2[Squ[2]][Squ[3]];
	      Squ[3] = 0;
	    }
	}
      fpos->Double = 0;
      for (i = 0; i < 4; i++)
	if (PI[i])
	  fpos->Double += fpos->m[i] * Squ[i];
    }
}

static char AMMET[64][64];
static void
ammet_iniz ()
{
  int sq, to;
  for (sq = A1; sq <= H8; sq++)
    for (to = A1; to <= H8; to++)
      {
	AMMET[sq][to] = (((RANK (to) > RANK (sq)) || (FILE (to) < FILE (sq)))
			 && (to - sq) != 10 && (to - sq) != -17);
	if ((RANK (to) - RANK (sq)) == (FILE (to) - FILE (sq))
	    && RANK (sq) > FILE (sq))
	  AMMET[sq][to] = 0;
      }
}


#define DIAGa1h8 (0x8040201008040201)
#define DIAGh1a8 (0x0102040810204080)

#include <time.h>
#include <pthread.h>
static void
nano (int ns)
{
  struct timespec TS[1];
  TS->tv_sec = 0;
  TS->tv_nsec = ns;
  nanosleep (TS, NULL);
}

typedef struct
{
  uint32 quo, rem;
} tipo_QUO_REM;

static tipo_QUO_REM **BUFFER, *BUF_DEST;
#define BUF_CONTO 4
static int buf_conto[BUF_CONTO], BUF_CHE;
static int BUF_DENSO[BUF_CONTO], BUF_LINDO[BUF_CONTO];
static int MARCO_FINE;
#define BUF_PROP 1048576

static int BUF_FINE, IO_FINE;
static uint32 DEST_LIB;
static uint8 *PILA_DEST[1806];
static uint8 *DESTRA[32];
static int DES_RICER[1806];
static uint8 oDES[1806], hDES[1806], fDES[1806], lDES[1806];

static int BUF_UTIL[1806][BUF_CONTO];

static int ott_dest (RobboTotalBase * tb, int x);
static int NUMERO_DI_ITERAZIONE = 0;

static uint8 *DEST_BITS[1806];
static uint8 *des_bit_pila[32];
static uint32 dbp_libero;
static int DES_NUM_FETTA[1806];

static int SPECCHIO[1806];
static void
iniz_iterazione ()
{
  int i, j;
  DEST_LIB = 0xffffffff;
  MARCO_FINE = FALSE;
  for (i = 0; i < 1806; i++)
    {
      PILA_DEST[i] = NULL;
      DES_RICER[i] = -1;
    }
  for (i = 0; i < 1806; i++)
    oDES[i] = hDES[i] = fDES[i] = lDES[i] = 0;
  for (i = 0; i < 1806; i++)
    SPECCHIO[i] = -1;
  for (i = 0; i < BUF_CONTO; i++)
    buf_conto[i] = BUF_DENSO[i] = 0;
  for (i = 0; i < BUF_CONTO; i++)
    BUF_LINDO[i] = 1;
  for (i = 0; i < 1806; i++)
    for (j = 0; j < BUF_CONTO; j++)
      BUF_UTIL[i][j] = 0;
  BUF_CHE = 0;
  BUF_DEST = BUFFER[BUF_CHE];
  BUF_LINDO[BUF_CHE] = 0;
}

static pthread_mutex_t PROIB_FSEEK[1];

void
iniz_una (RobboTotalBase * tb)
{
  int i;
  BUFFER = malloc (BUF_CONTO * sizeof (tipo_QUO_REM *));
  pthread_mutex_init (PROIB_FSEEK, NULL);
  for (i = 0; i < BUF_CONTO; i++)
    BUFFER[i] = malloc (BUF_PROP * sizeof (tipo_QUO_REM));
  for (i = 0; i < 32; i++)
    {
      DESTRA[i] = malloc (tb->PawnMult);
      des_bit_pila[i] = malloc (tb->PawnMult8);
    }
}

void
fine_una ()
{
  int i;
  for (i = 0; i < BUF_CONTO; i++)
    free (BUFFER[i]);
  free (BUFFER);
  for (i = 0; i < 32; i++)
    {
      free (DESTRA[i]);
      free (des_bit_pila[i]);
    }
}

static boolean
aMARCO (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    return emarco (tb->MAR,
		   fetta * tb->PawnMult8 * 8 + PositionNumber);
  return emarco (tb->mbuf, PositionNumber);
}

static void
aDESbits (RobboTotalBase * tb, int fetta, FILE * F, boolean w)
{
  if (!dbp_libero)
    {
      printf ("DESTRO BITS!\n");
      exit (1);
    }
  int r = BSF (dbp_libero);
  dbp_libero &= ~(1 << r);
  DES_NUM_FETTA[fetta] = r;
  fseek (F, fetta * tb->PawnMult8, 0);
  if (w)
    DISCO_FREAD2 (des_bit_pila[r], 1, tb->PawnMult8, F,
		  tb->per_pro[fetta]);
  else 
    DISCO_FREAD2 (des_bit_pila[r], 1, tb->PawnMult8, F,
		  tb->mar_pro[fetta]);
  DEST_BITS[fetta] = des_bit_pila[r];
}

static void
zDESbits (RobboTotalBase * tb, int fetta, FILE * F, boolean w)
{
  int r = DES_NUM_FETTA[fetta];
  if (r == -1)
    return;
  DES_NUM_FETTA[fetta] = -1;
  dbp_libero |= (1 << r);
  fseek (F, fetta * tb->PawnMult8, 0);
  int df = DISCO_FWRITE2 (des_bit_pila[r], 1, tb->PawnMult8, F);
  if (w) tb->per_pro[fetta] = df;
  else tb->mar_pro[fetta] = df;
  DEST_BITS[fetta] = NULL;
}

static void
MARCO (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    {
      Set (tb->MAR, fetta * tb->PawnMult8 * 8 + PositionNumber);
      return;
    }
  if (!DEST_BITS[fetta])
    aDESbits (tb, fetta, tb->Fmar, FALSE);
  Set (DEST_BITS[fetta], PositionNumber);
}

static void
nonMARCO (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    UnSet (tb->MAR, fetta * tb->PawnMult8 * 8 + PositionNumber);
  else
    UnSet (tb->mbuf, PositionNumber);
}

static boolean
eVIT (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    return emarco (tb->VIT, fetta * tb->PawnMult8 * 8 + PositionNumber);
  return emarco (tb->vbuf, PositionNumber);
}

static void
aVIT (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    Set (tb->VIT, fetta * tb->PawnMult8 * 8 + PositionNumber);
  else
    Set (tb->vbuf, PositionNumber);
}

static boolean
ePERD (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    return emarco (tb->PER, fetta * tb->PawnMult8 * 8 + PositionNumber);
  return emarco (tb->pbuf, PositionNumber);
}

static void
aPERD (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    {
      Set (tb->PER, fetta * tb->PawnMult8 * 8 + PositionNumber);
      return;
    }
  Set (DEST_BITS[fetta], PositionNumber);
}

static void
zPERD (RobboTotalBase * tb, int fetta, uint64 PositionNumber)
{
  if (tb->MEMORIA)
    UnSet (tb->PER, fetta * tb->PawnMult8 * 8 + PositionNumber);
  else
    UnSet (tb->pbuf, PositionNumber);
}

static void
buf_limp (tipo_QUO_REM * BUF_QUO_REM, int PROP, RobboTotalBase * tb)
{
  int b, Value, fetta, spos;
  for (b = 0; b < PROP; b++)
    {
      fetta = BUF_QUO_REM[b].quo;
      if (fetta == 0xbad)
	continue;
      spos = BUF_QUO_REM[b].rem;
      if (PILA_DEST[fetta] == NULL)
	while (!ott_dest (tb, fetta))
	  nano (1000);
      oDES[fetta] = 1;
      Value = PILA_DEST[fetta][spos];
      if (Value >= PRIMERA_PERDITA)
	{
	  printf ("MALATO z:%d %d/%d\n", Value, fetta, spos);
	  exit (1);
	}
      if (Value == 1)
	{
	  aPERD (tb, fetta, spos);
	  if (NUMERO_DI_ITERAZIONE >= LARGH)
	    {
	      uint64 n = (((uint64) fetta) << 32) | spos;
	      PILA_DEST[fetta][spos] = PERD_FINE;
	      tb->supp++;
	      fwrite (&n, 8, 1, tb->Fsupp);
	    }
	  else
	    PILA_DEST[fetta][spos] = PERD_IN (NUMERO_DI_ITERAZIONE);
	}
      else if (Value <= MASSIMO_RIMA)
	PILA_DEST[fetta][spos] = Value - 1;
    }
}

static void
buf_scan (RobboTotalBase * tb)
{
  int i, j, pro = tb->PEDONE ? 1806 : 462;
  while (!MARCO_FINE)
    {
      for (i = 0; i < BUF_CONTO; i++)
	if (BUF_DENSO[i])
	  {
	    buf_limp (BUFFER[i], buf_conto[i], tb);
	    buf_conto[i] = 0;
	    for (j = 0; j < pro; j++)
	      BUF_UTIL[j][i] = 0;
	    BUF_DENSO[i] = 0;
	    BUF_LINDO[i] = 1;
	  }
      if (tb->p[2])
	nano (10000);
    }
  if (tb->p[1])
    nano (10000);
  for (i = 0; i < BUF_CONTO; i++)
    if (BUF_DENSO[i])
      {
	buf_limp (BUFFER[i], buf_conto[i], tb);
	buf_conto[i] = 0;
	for (j = 0; j < pro; j++)
	  BUF_UTIL[j][i] = 0;
	BUF_DENSO[i] = 0;
	BUF_LINDO[i] = 1;
      }
  if (tb->p[1])
    nano (10000);
  for (i = 0; i < BUF_CONTO; i++)
    if (BUF_DENSO[i])
      {
	buf_limp (BUFFER[i], buf_conto[i], tb);
	buf_conto[i] = 0;
	for (j = 0; j < pro; j++)
	  BUF_UTIL[j][i] = 0;
	BUF_DENSO[i] = 0;
	BUF_LINDO[i] = 1;
      }
}

typedef struct
{
  RobboTotalBase *sinRTB, *desRTB;
} typeTotalBase2;

static void *
pthread_buf (void *x)
{
  typeTotalBase2 *u;
  u = (typeTotalBase2 *) x;
  buf_scan (u->desRTB);
  BUF_FINE = TRUE;
  return NULL;
}

#define LOCK(x) pthread_mutex_lock(x)
#define UNLOCK(x) pthread_mutex_unlock(x)

static int
ott_dest (RobboTotalBase * tb, int x)
{
  LOCK (PROIB_FSEEK);
  if (PILA_DEST[x] != NULL)
    {
      UNLOCK (PROIB_FSEEK);
      return 1;
    }
  if (!DEST_LIB)
    {
      UNLOCK (PROIB_FSEEK);
      return 0;
    }
  int r = BSF (DEST_LIB);
  DEST_LIB ^= (1 << r);
  aDESbits (tb, x, tb->Fper, TRUE);
  fseek (tb->Frima, x * tb->PawnMult, 0);
  DISCO_FREAD1 (DESTRA[r], 1, tb->PawnMult, tb->Frima, tb->rima_pro[x]);
  PILA_DEST[x] = DESTRA[r];
  DES_RICER[x] = r;
  lDES[x] = 0;
  hDES[x] = 1;
  UNLOCK (PROIB_FSEEK);
  return 3;
}

static int
DESlim (RobboTotalBase * tb, int x)
{
  LOCK (PROIB_FSEEK);
  int b, r = DES_RICER[x];
  uint8 LIMP = 0;
  if (r == -1)
    {
      lDES[x] = 1;
      UNLOCK (PROIB_FSEEK);
      return 1;
    }
  if (lDES[x])
    {
      UNLOCK (PROIB_FSEEK);
      return 2;
    }
  while (1)
    {
      LIMP = 0;
      for (b = 0; b < BUF_CONTO; b++)
	if (BUF_UTIL[x][b] && BUF_DENSO[b])
	  LIMP |= (1 << b);
      if (!LIMP)
	break;
      UNLOCK (PROIB_FSEEK);
      return 0;
    }
  if (SPECCHIO[x] == -1)
    {
      if (r == 0xadd)
	{
	  DES_RICER[x] = -1;
	}
      else
	{
	  SPECCHIO[x] = -1;
	  DEST_LIB ^= (1 << r);
	  DES_RICER[x] = -1;
	}
    }
  else
    {
      SPECCHIO[x] = -1;
      if (r == 0xadd)
	DES_RICER[x] = -1;
    }
  fseek (tb->Frima, x * tb->PawnMult, 0);
  int df =
    DISCO_FWRITE1 (PILA_DEST[x], 1, tb->PawnMult, tb->Frima);
  tb->rima_pro[x] = df;
  zDESbits (tb, x, tb->Fper, TRUE);
  PILA_DEST[x] = NULL;
  lDES[x] = 1;
  UNLOCK (PROIB_FSEEK);
  return 3;
}

static void
io_scan (RobboTotalBase * sinRTB, RobboTotalBase * desRTB)
{
  int i, pro = sinRTB->PEDONE ? 1806 : 462;
  while (!BUF_FINE)
    {
      for (i = 0; i < pro; i++)
	if (hDES[i] && fDES[i] && !lDES[i])
	  DESlim (desRTB, i);
      for (i = 0; i < pro; i++)
	if (oDES[i] && !hDES[i])
	  ott_dest (desRTB, i);
    }
  for (i = 0; i < pro; i++)
    if (fDES[i] && !lDES[i])
      DESlim (desRTB, i);
}

static void*
io_thread (void *x)
{
  typeTotalBase2 *u;
  u = (typeTotalBase2 *) x;
  io_scan (u->sinRTB, u->desRTB);
  IO_FINE = TRUE;
  fflush (u->sinRTB->Frima);
  fflush (u->desRTB->Frima);
  if (DEST_LIB != 0xffffffff)
    {
      printf ("MALATO DESTRO\n");
      exit (1);
    }
  return NULL;
}

static void
nuovo_iterazione (RobboTotalBase * sinRTB, RobboTotalBase * desRTB)
{
  iniz_iterazione ();
  pthread_t thr1[1], thr2[1];
  typeTotalBase2 u[1], v[1];
  v->sinRTB = u->sinRTB = sinRTB;
  v->desRTB = u->desRTB = desRTB;
  IO_FINE = FALSE;
  BUF_FINE = FALSE;
  pthread_create (thr1, NULL, io_thread, (void *) u);
  pthread_create (thr2, NULL, pthread_buf, (void *) v);
}

#define BUFDES(x, y) \
  { BUF_DEST[buf_conto[BUF_CHE]].quo = x; \
    BUF_DEST[buf_conto[BUF_CHE]++].rem = y; }

#define PREPDES(x) oDES[x] = 1;

#define PROSSIMO_BUF() \
  { BUF_DENSO[BUF_CHE] = 1; BUF_CHE += 1; BUF_CHE &= ( BUF_CONTO - 1 ); \
    while (!BUF_LINDO[BUF_CHE]) nano(1000); \
    BUF_LINDO[BUF_CHE] = 0; BUF_DEST = BUFFER[BUF_CHE]; }

#define DIAGa1h8SOPRA (0xff7f3f1f0f070301)
#define DIAGa1h8SOTTO (0x80c0e0f0f8fcfeff)
#define DIAGa1h8 (0x8040201008040201)

static void
sottRIMA (RobboTotalBase * desRTB, uint64 fet, uint64 ind)
{
  uint8 Value = desRTB->data[ind + fet * desRTB->PawnMult];
  if (Value >= PRIMERA_PERDITA)
    {
      printf ("MALATO z:%d %lld/%lld\n", Value, fet, ind);
      exit (1);
    }
  if (Value == 1)
    {
      aPERD (desRTB, fet, ind);
      if (NUMERO_DI_ITERAZIONE >= LARGH)
	{
	  uint64 n = (((uint64) fet) << 32) | ind;
	  desRTB->data[ind + fet * desRTB->PawnMult] = PERD_FINE;
	  desRTB->supp++;
	  fwrite (&n, 8, 1, desRTB->Fsupp);
	}
      else
	desRTB->data[ind + fet * desRTB->PawnMult] =
	  PERD_IN (NUMERO_DI_ITERAZIONE);
    }
  else if (Value <= MASSIMO_RIMA)
    desRTB->data[ind + fet * desRTB->PawnMult] = Value - 1;
}

static void
indice_sini (uint64 fet, uint64 ind, RobboTotalBase * sinRTB,
	     RobboTotalBase * desRTB)
{
  if (desRTB->MEMORIA)
    {
      sottRIMA (desRTB, fet, ind);
      return;
    }
  PREPDES (fet);
  BUFDES (fet, ind);
  BUF_UTIL[fet][BUF_CHE] = 1;
  if (buf_conto[BUF_CHE] == BUF_PROP)
    {
      PROSSIMO_BUF ();
    }
}

static void
indice_DES (uint64 fet, uint64 ind, RobboTotalBase * sinRTB,
	    RobboTotalBase * desRTB)
{
  MARCO (sinRTB, fet, ind);
}

static void
mossaPROCi (int sq, int b, int i, tipo_fPOS * fpos,
	    RobboTotalBase * sinRTB, RobboTotalBase * desRTB, boolean simm,
	    void (*INDICE) (uint64, uint64, RobboTotalBase *, RobboTotalBase *))
{
  uint64 index, Double;
  int to, a1, a2, a3;
  if (simm && !AMMET[sq][b])
    return;
  if (!fpos->rip[i])
    {
      index = fpos->ind + (b - sq) * fpos->m[i];
      Double = 0;
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
      if (!fpos->PEDONE && SqSet[fpos->wK] & DIAGa1h8
	  && SqSet[fpos->bK] & DIAGa1h8)
	{
	  Double =
	    fpos->Double + (lanc_norm[1][b] - lanc_norm[1][sq]) * fpos->m[i];
	  if (Double != index || !simm)
	    INDICE (fpos->re_fetta, Double, sinRTB, desRTB);
	}
    }
  if (fpos->rip[i] == 2)
    {
      if (fpos->rip_ai[i] == 1)
	to = 1;
      else
	to = -1;
      index = fpos->ind + fpos->m[i] * (STESSO2[b][fpos->sq[i + to]] -
					STESSO2[sq][fpos->sq[i + to]]);
      Double = 0;
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
      if (!fpos->PEDONE && SqSet[fpos->wK] & DIAGa1h8
	  && SqSet[fpos->bK] & DIAGa1h8)
	{
	  Double = fpos->Double + fpos->m[i] *
	    (STESSO2[lanc_norm[1][b]][lanc_norm[1][fpos->sq[i + to]]] -
	     STESSO2[lanc_norm[1][sq]][lanc_norm[1][fpos->sq[i + to]]]);
	  if (Double != index || !simm)
	    INDICE (fpos->re_fetta, Double, sinRTB, desRTB);
	}
    }
  if (fpos->rip[i] == 3)
    {
      if (fpos->rip_ai[i] == 1)
	{
	  a1 = 1;
	  a2 = 2;
	}
      if (fpos->rip_ai[i] == 2)
	{
	  a1 = -1;
	  a2 = 1;
	}
      if (fpos->rip_ai[i] == 3)
	{
	  a1 = -2;
	  a2 = -1;
	}
      index = fpos->ind +
	      fpos->m[i] * (STESSO3[b][fpos->sq[i + a1]][fpos->sq[i + a2]] -
			    STESSO3[sq][fpos->sq[i + a1]][fpos->sq[i + a2]]);
      Double = 0;
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
      if (!fpos->PEDONE && SqSet[fpos->wK] & DIAGa1h8
	  && SqSet[fpos->bK] & DIAGa1h8)
	{
	  Double = fpos->Double + fpos->m[i] *
	    (STESSO3[lanc_norm[1][b]][lanc_norm[1][fpos->sq[i + a1]]]
	     [lanc_norm[1][fpos->sq[i + a2]]] -
	     STESSO3[lanc_norm[1][sq]][lanc_norm[1][fpos->sq[i + a1]]]
	     [lanc_norm[1][fpos->sq[i + a2]]]);
	  if (Double != index || !simm)
	    INDICE (fpos->re_fetta, Double, sinRTB, desRTB);
	}
    }
  if (fpos->rip[i] == 4)
    {
      if (fpos->rip_ai[i] == 1)
	{
	  a1 = 1;
	  a2 = 2;
	  a3 = 3;
	}
      if (fpos->rip_ai[i] == 2)
	{
	  a1 = -1;
	  a2 = 1;
	  a3 = 2;
	}
      if (fpos->rip_ai[i] == 3)
	{
	  a1 = -2;
	  a2 = -1;
	  a3 = 1;
	}
      if (fpos->rip_ai[i] == 4)
	{
	  a1 = -3;
	  a2 = -2;
	  a3 = -1;
	}
      index = fpos->ind +
	fpos->m[i] *
	(STESSO4[b][fpos->sq[i + a1]][fpos->sq[i + a2]][fpos->sq[i + a3]] -
	 STESSO4[sq][fpos->sq[i + a1]][fpos->sq[i + a2]][fpos->sq[i + a3]]);
      Double = 0;
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
      if (!fpos->PEDONE && SqSet[fpos->wK] & DIAGa1h8 && SqSet[fpos->bK] & DIAGa1h8)
	{
	  Double = fpos->Double + fpos->m[i] *
	    (STESSO4[lanc_norm[1][b]][lanc_norm[1][fpos->sq[i + a1]]]
	     [lanc_norm[1][fpos->sq[i + a2]]][lanc_norm[1][fpos->sq[i + a3]]] -
	     STESSO4[lanc_norm[1][sq]][lanc_norm[1][fpos->sq[i + a1]]]
	     [lanc_norm[1][fpos->sq[i + a2]]][lanc_norm[1][fpos->sq[i + a3]]]);
	  if (Double != index || !simm)
	    INDICE (fpos->re_fetta, Double, sinRTB, desRTB);
	}
    }
}

static void
iMOSSA_PEDONE (int sq, int b, int i, tipo_fPOS * fpos,
	       RobboTotalBase * sinRTB, RobboTotalBase * desRTB,
	       void (*INDICE) (uint64, uint64, RobboTotalBase *,
			       RobboTotalBase *))
{
  int to, a1, a2, a3;
  uint64 index;
  if (!fpos->rip[i])
    {
      index = fpos->ind + (b - sq) * fpos->m[i];
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
    }
  if (fpos->rip[i] == 2)
    {
      if (fpos->rip_ai[i] == 1)
	to = 1;
      else
	to = -1;
      index = fpos->ind + fpos->m[i] * (STESSO2[b - 8][fpos->sq[i + to] - 8] -
					STESSO2[sq - 8][fpos->sq[i + to] - 8]);
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
    }
  if (fpos->rip[i] == 3)
    {
      if (fpos->rip_ai[i] == 1)
	{
	  a1 = 1;
	  a2 = 2;
	}
      if (fpos->rip_ai[i] == 2)
	{
	  a1 = -1;
	  a2 = 1;
	}
      if (fpos->rip_ai[i] == 3)
	{
	  a1 = -2;
	  a2 = -1;
	}
      index = fpos->ind +
	fpos->m[i] *
	(STESSO3[(b) - 8][fpos->sq[i + a1] - 8][fpos->sq[i + a2] - 8] -
	 STESSO3[(sq) - 8][fpos->sq[i + a1] - 8][fpos->sq[i + a2] - 8]);
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
    }
  if (fpos->rip[i] == 4)
    {
      if (fpos->rip_ai[i] == 1)
	{
	  a1 = 1;
	  a2 = 2;
	  a3 = 3;
	}
      if (fpos->rip_ai[i] == 2)
	{
	  a1 = -1;
	  a2 = 1;
	  a3 = 2;
	}
      if (fpos->rip_ai[i] == 3)
	{
	  a1 = -2;
	  a2 = -1;
	  a3 = 1;
	}
      if (fpos->rip_ai[i] == 4)
	{
	  a1 = -3;
	  a2 = -2;
	  a3 = -1;
	}
      index = fpos->ind +
	fpos->m[i] *
	(STESSO4[(b) - 8][fpos->sq[i + a1] - 8]
	 [fpos->sq[i + a2] - 8][fpos->sq[i + a3] - 8] -
	 STESSO4[(sq) - 8][fpos->sq[i + a1] - 8]
	 [fpos->sq[i + a2] - 8][fpos->sq[i + a3] - 8]);
      INDICE (fpos->re_fetta, index, sinRTB, desRTB);
    }
}

static void
iMOSSA_RE (uint64 TARGA, tipo_fPOS * fpos,
	   RobboTotalBase * sinRTB, RobboTotalBase * desRTB, boolean simm,
	   void (*INDICE) (uint64, uint64, RobboTotalBase *,
			   RobboTotalBase *))
{
  int j;
  uint64 Double, index;
  while (TARGA)
    {
      int b = BSF (TARGA);
      BitClear (b, TARGA);
      Double = 0;
      if (fpos->PEDONE)
	{
	  INDICE (RE_pedone[64 * b + fpos->bK], fpos->ind, sinRTB, desRTB);
	}
      else
	{
	  if (simm && !AMMET[fpos->wK][b])
	    continue;
	  if (SqSet[fpos->bK] & DIAGa1h8 && SqSet[b] & DIAGa1h8SOPRA)
	    {
	      if (SqSet[b] & DIAGa1h8)
		{
		  INDICE (RE_normale[64 * b + fpos->bK], fpos->ind, sinRTB, desRTB);
		  for (j = 0; j < 4; j++)
		    if (fpos->pi[j])
		      {
			if (!fpos->rip[j])
			  Double += fpos->m[j] * lanc_norm[1][fpos->sq[j]];
			if (fpos->rip[j] == 2 && fpos->rip_ai[j] == 2)
			  Double += fpos->m[j] *
			              STESSO2[lanc_norm[1][fpos->sq[j]]]
			                     [lanc_norm[1][fpos->sq[j - 1]]];
			if (fpos->rip[j] == 3 && fpos->rip_ai[j] == 3)
			  Double += fpos->m[j] *
			              STESSO3[lanc_norm[1][fpos->sq[j]]]
			                     [lanc_norm[1][fpos->sq[j - 1]]]
			                     [lanc_norm[1][fpos->sq[j - 2]]];
			if (fpos->rip[j] == 4 && fpos->rip_ai[j] == 4)
			  Double += fpos->m[j] *
			              STESSO4[lanc_norm[1][fpos->sq[j]]]
			                     [lanc_norm[1][fpos->sq[j - 1]]]
			                     [lanc_norm[1][fpos->sq[j - 2]]]
			                     [lanc_norm[1][fpos->sq[j - 3]]];
		      }
		  if (Double != fpos->ind || !simm)
		    INDICE (RE_normale[64 * b + fpos->bK], Double, sinRTB,
			    desRTB);
		}
	      else
		{
		  index = 0;
		  for (j = 0; j < 4; j++)
		    if (fpos->pi[j])
		      {
			if (!fpos->rip[j])
			  index += fpos->m[j] * lanc_norm[1][fpos->sq[j]];
			if (fpos->rip[j] == 2 && fpos->rip_ai[j] == 2)
			  index += fpos->m[j] *
			    STESSO2[lanc_norm[1][fpos->sq[j]]]
			            [lanc_norm[1][fpos->sq[j - 1]]];
			if (fpos->rip[j] == 3 && fpos->rip_ai[j] == 3)
			  index += fpos->m[j] *
			    STESSO3[lanc_norm[1][fpos->sq[j]]]
			           [lanc_norm[1][fpos->sq[j - 1]]]
			           [lanc_norm[1][fpos->sq[j - 2]]];
			if (fpos->rip[j] == 4 && fpos->rip_ai[j] == 4)
			  index += fpos->m[j] *
			    STESSO4[lanc_norm[1][fpos->sq[j]]]
			           [lanc_norm[1][fpos->sq[j - 1]]]
			           [lanc_norm[1][fpos->sq[j - 2]]]
			           [lanc_norm[1][fpos->sq[j - 3]]];
		      }
		  INDICE (RE_normale
			  [64 * lanc_norm[1][b] + lanc_norm[1][fpos->bK]],
			  index, sinRTB, desRTB);
		}
	    }
	  else
	    {
	      INDICE (RE_normale[64 * b + fpos->bK], fpos->ind, sinRTB,
		      desRTB);
	    }
	}
    }
}

#define mossaPROC(sq, b) \
 mossaPROCi(sq, b, i, fpos, Tot1, Tot2, simm, indice_sini)
#define MOSSA_PEDONE(sq, b) \
 iMOSSA_PEDONE(sq, b, i, fpos, Tot1, Tot2, indice_sini)
#define MOSSA_RE(TARGA) \
 iMOSSA_RE(TARGA, fpos, Tot1, Tot2, simm, indice_sini)

#define LOOP(TARGA, pros_bit, O) \
  { uint64 T = TARGA; \
    while (T) \
      { int b = pros_bit(T); \
	if (SqSet[b] & O) break; \
	T &= ~SqSet[b]; \
	mossaPROC(sq,b); } }

#define eSTRUTvalu(A,x) A[x]
#define aSTRUTvalu(A,x,y) A[x] = y

static sint16 FINE_DES_NORMALE[462][8], FINE_DES_PEDONE[1806][8];
static uint16 RIMAP_NORMALE[462], RIMAP_PEDONE[1806];

static sint64
processo_MARCO (RobboTotalBase * Tot1, int num, RobboTotalBase * Tot2)
{
  sint64 count = 0;
  int i, j;
  boolean simm;
  uint64 Clock = GetClock ();
  tipo_fPOS fpos[1];
  uint64 rf, refet, RF = Tot1->PEDONE ? 1806 : 462;
  uint64 PositionNumber, LIMITE = Tot1->PawnMult;
  printf ("Mossa %d\n", num);
  for (i = 0; i < RF; i++)
    {
      DEST_BITS[i] = NULL;
      DES_NUM_FETTA[i] = -1;
    }
  dbp_libero = 0xffffffff;
  for (refet = 0; refet < RF; refet++)
    {
      rf = Tot1->PEDONE ? RIMAP_PEDONE[refet] : RIMAP_NORMALE[refet];
      if (!Tot1->MEMORIA)
	{
	  fseek (Tot1->Fmar, rf * Tot1->PawnMult8, 0);
	  DISCO_FREAD2 (Tot1->mbuf, 1, Tot1->PawnMult8, Tot1->Fmar,
			Tot1->mar_pro[rf]);
	  fseek (Tot1->Fvit, rf * Tot1->PawnMult8, 0);
	  DISCO_FREAD2 (Tot1->vbuf, 1, Tot1->PawnMult8, Tot1->Fvit,
			Tot1->vit_pro[rf]);
	}
      for (PositionNumber = 0; PositionNumber < LIMITE; PositionNumber++)
	{
	  if (!aMARCO (Tot1, rf, PositionNumber))
	    continue;
	  if (eVIT (Tot1, rf, PositionNumber))
	    {
	      nonMARCO (Tot1, rf, PositionNumber);
	      continue;
	    }
	  aVIT (Tot1, rf, PositionNumber);
	  FORZA_POS (Tot1, rf * LIMITE + PositionNumber, fpos);
#if 0
	  printf("VITT %d %d/%d rf:%d wK:%d bK:%d %d@%d %d@%d %d@%d %llx %d\n",
		 rf, PositionNumber, LIMITE,
		 fpos->re_fetta, fpos->wK, fpos->bK,
		 fpos->pi[0], fpos->sq[0],
		 fpos->pi[1], fpos->sq[1],
		 fpos->pi[2], fpos->sq[2],
		 fpos->Occupied, fpos->ind);
#endif
	  uint64 Occupied = fpos->Occupied;
	  int sq, pi;
	  nonMARCO (Tot1, rf, PositionNumber);
	  simm = FALSE;
	  if (fpos->Double != -1)
	    {
	      if (fpos->ind < fpos->Double)
		continue;
	      if (fpos->ind == fpos->Double)
		simm = TRUE;
	    }
	  count++;
	  for (i = 0; i < 4 && fpos->pi[i]; i++)
	    {
	      if (fpos->pi[i] >= 8)
		continue;
	      sq = fpos->sq[i];
	      pi = fpos->pi[i];
	      if (pi == wEnumP)
		{
		  if (sq >= A3 && !(Occupied & SqSet[sq - 8]))
		    {
		      MOSSA_PEDONE (sq, sq - 8);
		      if (RANK (sq) == R4 && !(Occupied & SqSet[sq - 16]))
			{
			  if (FILE (sq) != FA
			      && fpos->PawnOnFourth & SqSet[sq - 1])
			    {
			      type_PiSq PiSq[1];
			      int w = 0;
			      PiSq->wK = fpos->wK;
			      PiSq->bK = fpos->bK;
			      PiSq->wtm = TRUE;
			      PiSq->pedone = TRUE;
			      for (j = 0; j < 4; j++)
				{
				  if (fpos->sq[j] == (sq - 1))
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = sq - 8;
				    }
				  else if (fpos->sq[j] == (sq - 9) &&
					   fpos->pi[j] == BlockedPawn)
				    {
				      PiSq->pi[w] = bEnumP;
				      PiSq->sq[w++] = sq - 8;
				      PiSq->pi[w] = wEnumP;
				      PiSq->sq[w++] = sq - 9;
				    }
				  else if (fpos->sq[j] != sq)
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = fpos->sq[j];
				    }
				}
			      for (j = w; j < 4; j++)
				PiSq->pi[j] = PiSq->sq[j] = 0;
			      int tb, rf;
			      uint64 ind, Double;
			      tot_ind_ott (PiSq, &tb, &ind, &Double, &rf);
			      if (DISCO_PERDITA
				  (tot_valu
				   (TABLE_TOTAL_BASES + tb,
				    rf * (TABLE_TOTAL_BASES + tb)->PawnMult + ind)))
				continue;
			    }
			  if (FILE (sq) != FH
			      && fpos->PawnOnFourth & SqSet[sq + 1])
			    {
			      type_PiSq PiSq[1];
			      int w = 0;
			      PiSq->wK = fpos->wK;
			      PiSq->bK = fpos->bK;
			      PiSq->wtm = TRUE;
			      PiSq->pedone = TRUE;
			      for (j = 0; j < 4; j++)
				{
				  if (fpos->sq[j] == (sq + 1))
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = sq - 8;
				    }
				  else if (fpos->sq[j] == (sq - 7) &&
					   fpos->pi[j] == BlockedPawn)
				    {
				      PiSq->pi[w] = bEnumP;
				      PiSq->sq[w++] = sq - 8;
				      PiSq->pi[w] = wEnumP;
				      PiSq->sq[w++] = sq - 7;
				    }
				  else if (fpos->sq[j] != sq)
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = fpos->sq[j];
				    }
				}
			      for (j = w; j < 4; j++)
				PiSq->pi[j] = PiSq->sq[j] = 0;
			      int tb, rf;
			      uint64 ind, Double;
			      tot_ind_ott (PiSq, &tb, &ind, &Double, &rf);
			      if (DISCO_PERDITA
				  (tot_valu
				   (TABLE_TOTAL_BASES + tb,
				    rf * (TABLE_TOTAL_BASES + tb)->PawnMult + ind)))
				continue;
			    }
			  MOSSA_PEDONE (sq, sq - 16);
			}
		    }
		}
	      if (pi == wEnumN)
		LOOP (AttN[sq] & ~fpos->Occupied, BSF, 0);
	      if (pi == wEnumB || pi == wEnumQ)
		{
		  LOOP (RAGGIO_NO[sq], BSF, Occupied);
		  if (!simm)
		    LOOP (RAGGIO_SE[sq], BSR, Occupied);
		  if (!simm || SqSet[sq] & DIAGa1h8SOTTO)
		    {
		      LOOP (RAGGIO_SO[sq], BSR, Occupied);
		      LOOP (RAGGIO_NE[sq], BSF, Occupied);
		    }
		}
	      if (pi == wEnumR || pi == wEnumQ)
		{
		  LOOP (RAGGIO_N[sq], BSF, Occupied);
		  LOOP (RAGGIO_O[sq], BSR, Occupied);
		  if (!simm)
		    {
		      LOOP (RAGGIO_S[sq], BSR, Occupied);
		      LOOP (RAGGIO_E[sq], BSF, Occupied);
		    }
		}
	    }
	  uint64 RE_TARGA = AttK[fpos->wK] & ~AttK[fpos->bK] & ~Occupied;
	  MOSSA_RE (RE_TARGA);
	}
      if (Tot1->MEMORIA)
	continue;
      PROSSIMO_BUF ();
      int df;
      fseek (Tot1->Fmar, rf * Tot1->PawnMult8, 0);
      df = DISCO_FWRITE2 (Tot1->mbuf, 1, Tot1->PawnMult8, Tot1->Fmar);
      Tot1->mar_pro[rf] = df;
      fseek (Tot1->Fvit, rf * Tot1->PawnMult8, 0);
      df = DISCO_FWRITE2 (Tot1->vbuf, 1, Tot1->PawnMult8, Tot1->Fvit);
      Tot1->vit_pro[rf] = df;
      if (Tot1->PEDONE)
	for (i = 0; FINE_DES_PEDONE[refet][i] != -1; i++)
	  fDES[FINE_DES_PEDONE[refet][i]] = 1;
      else
	for (i = 0; FINE_DES_NORMALE[refet][i] != -1; i++)
	  fDES[FINE_DES_NORMALE[refet][i]] = 1;
    }
  printf ("MARCO %lld [%.2fs]\n", count,
	  ((double) (GetClock () - Clock)) / 1000000.0);
  if (Tot1->MEMORIA)
    return count;
  return count;
}

#define MOSSA_PEDONE(sq, b) \
  iMOSSA_PEDONE(sq, b, i, fpos, Tot2, Tot1, indice_DES)
#define MOSSA_RE(TARGA) \
  iMOSSA_RE(TARGA, fpos, Tot2, Tot1, simm, indice_DES)
#define mossaPROC(sq, b) \
  mossaPROCi(sq, b, i, fpos, Tot2, Tot1, simm, indice_DES)

static sint64
procRIMA (RobboTotalBase * Tot1, RobboTotalBase * Tot2)
{
  sint64 count = 0;
  int i, j;
  uint64 PositionNumber, Clock = GetClock ();
  tipo_fPOS fpos[1];
  int refet, rf, RF = Tot1->PEDONE ? 1806 : 462;
  for (i = 0; i < RF; i++)
    {
      DEST_BITS[i] = NULL;
      DES_NUM_FETTA[i] = -1;
    }
  dbp_libero = 0xffffffff;
  for (refet = 0; refet < RF; refet++)
    {
      rf = Tot1->PEDONE ? RIMAP_PEDONE[refet] : RIMAP_NORMALE[refet];
      if (!Tot1->MEMORIA)
	{
	  fseek (Tot1->Fper, rf * Tot1->PawnMult8, 0);
	  DISCO_FREAD2 (Tot1->pbuf, 1, Tot1->PawnMult8, Tot1->Fper,
			Tot1->per_pro[rf]);
	}
      for (PositionNumber = 0; PositionNumber < Tot1->PawnMult; PositionNumber++)
	{
	  if (!ePERD (Tot1, rf, PositionNumber))
	    continue;
	  zPERD (Tot1, rf, PositionNumber);
	  FORZA_POS (Tot1, rf * Tot1->PawnMult + PositionNumber, fpos);
#if 0
	  printf("PERD %d %d/%d rf:%d wK:%d bK:%d %d@%d %d@%d %d@%d %llx %d\n",
		 rf, PositionNumber, Tot1->PawnMult,
		 fpos->re_fetta, fpos->wK, fpos->bK,
		 fpos->pi[0], fpos->sq[0],
		 fpos->pi[1], fpos->sq[1],
		 fpos->pi[2], fpos->sq[2],
		 fpos->Occupied, fpos->ind);
#endif
	  uint64 Occupied = fpos->Occupied;
	  int sq, pi;
	  boolean simm = FALSE;
	  if (fpos->Double != -1)
	    {
	      if (fpos->ind < fpos->Double)
		continue;
	      if (fpos->ind == fpos->Double)
		simm = TRUE;
	    }
	  count++;
	  for (i = 0; i < 4 && fpos->pi[i]; i++)
	    {
	      if (fpos->pi[i] >= 8)
		continue;
	      sq = fpos->sq[i];
	      pi = fpos->pi[i];
	      if (pi == wEnumP)
		{
		  if (sq >= A3 && !(Occupied & SqSet[sq - 8]))
		    {
		      MOSSA_PEDONE (sq, sq - 8);
		      if (RANK (sq) == R4 && !(Occupied & SqSet[sq - 16]))
			{
			  if (FILE (sq) != FA
			      && fpos->PawnOnFourth & SqSet[sq - 1])
			    {
			      type_PiSq PiSq[1];
			      int w = 0;
			      PiSq->wK = fpos->wK;
			      PiSq->bK = fpos->bK;
			      PiSq->wtm = TRUE;
			      PiSq->pedone = TRUE;
			      for (j = 0; j < 4; j++)
				{
				  if (fpos->sq[j] == (sq - 1))
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = sq - 8;
				    }
				  else if (fpos->sq[j] == (sq - 9) &&
					   fpos->pi[j] == BlockedPawn)
				    {
				      PiSq->pi[w] = bEnumP;
				      PiSq->sq[w++] = sq - 8;
				      PiSq->pi[w] = wEnumP;
				      PiSq->sq[w++] = sq - 9;
				    }
				  else if (fpos->sq[j] != sq)
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = fpos->sq[j];
				    }
				}
			      for (j = w; j < 4; j++)
				PiSq->pi[j] = PiSq->sq[j] = 0;
			      int tb, rf;
			      uint64 ind, Double;
			      tot_ind_ott (PiSq, &tb, &ind, &Double, &rf);
			      if (DISCO_PATT_PERD
				  (tot_valu
				   ((TABLE_TOTAL_BASES + tb),
				    rf * (TABLE_TOTAL_BASES + tb)->PawnMult + ind)))
				continue;
			    }
			  if (FILE (sq) != FH
			      && fpos->PawnOnFourth & SqSet[sq + 1])
			    {
			      type_PiSq PiSq[1];
			      int w = 0;
			      PiSq->wK = fpos->wK;
			      PiSq->bK = fpos->bK;
			      PiSq->wtm = TRUE;
			      PiSq->pedone = TRUE;
			      for (j = 0; j < 4; j++)
				{
				  if (fpos->sq[j] == (sq + 1))
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = sq - 8;
				    }
				  else if (fpos->sq[j] == (sq - 7) &&
					   fpos->pi[j] == BlockedPawn)
				    {
				      PiSq->pi[w] = bEnumP;
				      PiSq->sq[w++] = sq - 8;
				      PiSq->pi[w] = wEnumP;
				      PiSq->sq[w++] = sq - 7;
				    }
				  else if (fpos->sq[j] != sq)
				    {
				      PiSq->pi[w] = fpos->pi[j];
				      PiSq->sq[w++] = fpos->sq[j];
				    }
				}
			      for (j = w; j < 4; j++)
				PiSq->pi[j] = PiSq->sq[j] = 0;
			      int tb, rf;
			      uint64 ind, Double;
			      tot_ind_ott (PiSq, &tb, &ind, &Double, &rf);
			      if (DISCO_PATT_PERD
				  (tot_valu
				   (TABLE_TOTAL_BASES + tb,
				    rf * (TABLE_TOTAL_BASES + tb)->PawnMult + ind)))
				continue;
			    }
			  MOSSA_PEDONE (sq, sq - 16);
			}
		    }
		}
	      if (pi == wEnumN)
		LOOP (AttN[sq] & ~fpos->Occupied, BSF, 0);
	      if (pi == wEnumB || pi == wEnumQ)
		{
		  LOOP (RAGGIO_NO[sq], BSF, Occupied);
		  if (!simm)
		    LOOP (RAGGIO_SE[sq], BSR, Occupied);
		  if (!simm || SqSet[sq] & DIAGa1h8SOTTO)
		    {
		      LOOP (RAGGIO_SO[sq], BSR, Occupied);
		      LOOP (RAGGIO_NE[sq], BSF, Occupied);
		    }
		}
	      if (pi == wEnumR || pi == wEnumQ)
		{
		  LOOP (RAGGIO_N[sq], BSF, Occupied);
		  LOOP (RAGGIO_O[sq], BSR, Occupied);
		  if (!simm)
		    {
		      LOOP (RAGGIO_S[sq], BSR, Occupied);
		      LOOP (RAGGIO_E[sq], BSF, Occupied);
		    }
		}
	    }
	  uint64 RE_TARGA = AttK[fpos->wK] & ~AttK[fpos->bK] & ~Occupied;
	  MOSSA_RE (RE_TARGA);
	}
      if (Tot1->MEMORIA)
	continue;
      fseek (Tot1->Fper, rf * Tot1->PawnMult8, 0);
      int df = DISCO_FWRITE2 (Tot1->pbuf, 1, Tot1->PawnMult8, Tot1->Fper);
      Tot1->per_pro[rf] = df;
      if (Tot1->PEDONE)
	for (i = 0; FINE_DES_PEDONE[refet][i] != -1; i++)
	  zDESbits (Tot2, FINE_DES_PEDONE[refet][i], Tot2->Fmar, FALSE);
      else
	for (i = 0; FINE_DES_NORMALE[refet][i] != -1; i++)
	  zDESbits (Tot2, FINE_DES_NORMALE[refet][i], Tot2->Fmar, FALSE);
    }
  printf ("NON RIMA %lld [%.2fs]\n", count,
	  ((double) (GetClock () - Clock)) / 1000000.0);
  return count;
}

uint64
fareMARCO (RobboTotalBase * Tot1, int iter, RobboTotalBase * Tot2)
{
  if (!Tot1->MEMORIA)
    nuovo_iterazione (Tot1, Tot2);
  NUMERO_DI_ITERAZIONE = iter + 1;
  if (NUMERO_DI_ITERAZIONE == LARGH)
    {
      char FN[1024];
      sprintf (FN, "%sSUPP.%s", Tot2->DIR_NOME, Tot2->string);
      Tot2->Fsupp = fopen (FN, "w+");
    }
  if (NUMERO_DI_ITERAZIONE >= LARGH)
    {
      fputc (0xff, Tot2->Fsupp);
      fputc (0xff, Tot2->Fsupp);
      fputc (0xff, Tot2->Fsupp);
      fputc (0xff, Tot2->Fsupp);
      fputc (NUMERO_DI_ITERAZIONE & 255, Tot2->Fsupp);
      fputc ((NUMERO_DI_ITERAZIONE >> 8) & 255, Tot2->Fsupp);
      fputc (0xff, Tot2->Fsupp);
      fputc (0xff, Tot2->Fsupp);
    }
  uint64 marco = processo_MARCO (Tot1, iter, Tot2);
  if (Tot1->MEMORIA)
    return marco;
  MARCO_FINE = TRUE;
  while (!IO_FINE || !BUF_FINE)
    nano (1000);
  return marco;
}

uint64
fareRIMA (RobboTotalBase * Tot1, int iter, RobboTotalBase * Tot2)
{
  return procRIMA (Tot1, Tot2);
}

typedef struct
{
  int w, max, min;
  sint16 data[10];
} typeKingPawn;

static int
qsort_RE_PEDONE (const void *x, const void *y)
{
  typeKingPawn *a, *b;
  a = (typeKingPawn *) x;
  b = (typeKingPawn *) y;
  if (a->min < b->min)
    return -1;
  if (a->min > b->min)
    return 1;
  if (a->w < b->w)
    return -1;
  if (a->w > b->w)
    return 1;
  return 0;
}

static void
iniz_fine_des ()
{
  int c, j, wK, bK, b, rf, i, t, w, max, min, FINE[1806];
  typeKingPawn RE_PEDONE[1806], RE_NORMALE[462];
  for (wK = A1; wK <= H8; wK++)
    for (bK = A1; bK <= H8; bK++)
      {
	rf = RE_pedone[64 * wK + bK];
	if (rf == -1)
	  continue;
	uint64 U = (AttK[bK] & ~AttK[wK]) | SqSet[bK];
	c = 0;
	max = 0;
	min = 2000;
	while (U)
	  {
	    b = BSF (U);
	    BitClear (b, U);
	    w = 1 + 2 * (FILE (wK) >= FE);
	    t = RE_pedone[lanc_pedone[w][b] * 64 + lanc_pedone[w][wK]];
	    if (t < min)
	      min = t;
	    if (t > max)
	      max = t;
	    RE_PEDONE[rf].data[c++] = t;
	  }
	RE_PEDONE[rf].max = max;
	RE_PEDONE[rf].min = min;
	RE_PEDONE[rf].w = rf;
	RE_PEDONE[rf].data[c] = -1;
      }
  qsort (RE_PEDONE, 1806, sizeof (typeKingPawn), qsort_RE_PEDONE);
  for (i = 0; i < 1806; i++)
    RIMAP_PEDONE[i] = RE_PEDONE[i].w;
  for (i = 0; i < 1806; i++)
    for (j = 0; RE_PEDONE[i].data[j] != -1; j++)
      FINE[RE_PEDONE[i].data[j]] = i;
  for (i = 0; i < 1806; i++)
    for (j = 0; j < 8; j++)
      FINE_DES_PEDONE[i][j] = -1;
  for (i = 0; i < 1806; i++)
    {
      int u = FINE[i];
      for (j = 0; FINE_DES_PEDONE[u][j] != -1; j++);
      FINE_DES_PEDONE[u][j] = i;
    }

  for (wK = A1; wK <= H8; wK++)
    for (bK = A1; bK <= H8; bK++)
      {
	uint64 U = (AttK[bK] & ~AttK[wK]) | SqSet[bK];
	rf = RE_normale[64 * wK + bK];
	if (rf == -1)
	  continue;
	c = 0;
	max = 0;
	min = 500;
	while (U)
	  {
	    b = BSF (U);
	    BitClear (b, U);
	    w = oLANC[b][wK];
	    t = RE_normale[lanc_norm[w][b] * 64 + lanc_norm[w][wK]];
	    if (t < min)
	      min = t;
	    if (t > max)
	      max = t;
	    RE_NORMALE[rf].data[c++] = t;
	  }
	RE_NORMALE[rf].max = max;
	RE_NORMALE[rf].min = min;
	RE_NORMALE[rf].w = rf;
	RE_NORMALE[rf].data[c] = -1;
      }
  qsort (RE_NORMALE, 462, sizeof (typeKingPawn), qsort_RE_PEDONE);
  for (i = 0; i < 462; i++)
    RIMAP_NORMALE[i] = RE_NORMALE[i].w;
  for (i = 0; i < 462; i++)
    for (j = 0; RE_NORMALE[i].data[j] != -1; j++)
      FINE[RE_NORMALE[i].data[j]] = i;
  for (i = 0; i < 462; i++)
    for (j = 0; j < 8; j++)
      FINE_DES_NORMALE[i][j] = -1;
  for (i = 0; i < 462; i++)
    {
      int u = FINE[i];
      for (j = 0; FINE_DES_NORMALE[u][j] != -1; j++);
      FINE_DES_NORMALE[u][j] = i;
    }
}

void
iniz_vario ()
{
  iniz_raggio ();
  ammet_iniz ();
  iniz_fine_des ();
}
