
#include "RobboLito.h"
#include <string.h>

static uint64 PV_HASH_SIZE = 0x10000;
static boolean PV_FLAG_HASH_INIT = FALSE;

void PVHashClear ()
{
  memset (PVHashTable, 0, PV_HASH_SIZE * sizeof (typePVHash));
}

int InitPVHash(int mb)
{
  if (mb > 1024)
    mb = 1024;
  if (mb < 1)
    mb = 1;
  PV_HASH_SIZE = ((1ULL << BSR (mb)) << 20) / sizeof (typePVHash);
  mb = (PV_HASH_SIZE * sizeof (typeHash)) >> 20;
  PVHashMask = PV_HASH_SIZE - 4;
  if (PV_FLAG_HASH_INIT)
    ALIGNED_FREE (PVHashTable);
  PV_FLAG_HASH_INIT = TRUE;
  MEMALIGN (PVHashTable, 64, PV_HASH_SIZE * sizeof (typePVHash));
  PVHashClear ();
  return mb;
}

/******************************** INIT ********************************/
#define MAX_AGE 256
#define MAX_DEPTH 256

#define AGE_DEPTH_MIX(x, y) \
  (((AGE - (x)) & (MAX_AGE - 1)) * MAX_DEPTH + (MAX_DEPTH - ((y) + 1)))

void IncrementAge ()
{
  AGE += 1;
  if (AGE == MAX_AGE)
    AGE = 0;
}

static uint64 HASH_SIZE;
static boolean FLAG_HASH_INIT = FALSE;

void HashClear ()
{
#ifndef HYPER_HASH
  int i;
  memset (HashTable, 0, HASH_SIZE * sizeof (typeHash));
  for (i = 0; i < HASH_SIZE; i++)
    (HashTable + i)->age = (MAX_AGE / 2);
  AGE = 0;
#else
  int i;
  uint64 j; /* 64-BIT */
  for (i = 0; i < 3; i++)
    if (TABLE_ARR[i])
      {
	memset (TABLE_ARR[i], 0, (MASK_ARR[i] + 4) * sizeof (typeHash));
	for (j = 0; j < MASK_ARR[i] + 4; j++)
	  (TABLE_ARR[i] + j)-> age = (MAX_AGE / 2);
      }
  AGE = 0;
#endif
}

int InitHash (int mb)
{
#ifndef HYPER_HASH
  AGE = 0;
  HASH_SIZE = ((1ULL << BSR (mb)) << 20) / sizeof (typeHash);
  if (HASH_SIZE > 0x100000000)
    HASH_SIZE = 0x100000000;
  mb = (HASH_SIZE * sizeof (typeHash)) >> 20;
  HashMask = HASH_SIZE - 4;
  if (FLAG_HASH_INIT)
    ALIGNED_FREE (HashTable);
  FLAG_HASH_INIT = TRUE;
  MEMALIGN (HashTable, 64, HASH_SIZE * sizeof (typeHash));
  HashClear ();
  return mb;
#else
  int i, w = BSR (mb), MB;
  int b0, b1, b2;
  AGE = 0;
  if (FLAG_HASH_INIT)
    {
      for (i = 0; i < 3; i++)
	if (TABLE_ARR[i])
	  ALIGNED_FREE (TABLE_ARR[i]);
    }
  FLAG_HASH_INIT = TRUE;
  b0 = b1 = b2 = 0;
  TABLE_ARR[0] = TABLE_ARR[1] = TABLE_ARR[2] = NULL;
  MASK_ARR[0] = MASK_ARR[1] = MASK_ARR[2] = 0;
  for (i = 0; i < 16; i++)
    HASH_WHICH[i] = 0;
  if (w >= 2)
    mb = (mb >> (w - 2)) << (w - 2);
  MB = mb;
  b0 = BSR (mb);
  MASK_ARR[0] = ((1ULL << (20 + b0)) / sizeof (typeHash)) - 4;
  mb -= (1 << b0);
  if (mb)
    {
      b1 = BSR (mb);
      MASK_ARR[1] = ((1ULL << (20 + b1)) / sizeof (typeHash)) - 4;
      mb -= (1 << b1);
    }
  if (mb)
    {
      b2 = BSR (mb);
      MASK_ARR[2] = ((1ULL << (20 + b2)) / sizeof (typeHash)) - 4;
      mb -= (1 << b2);
    }
#define WHICH_SPLIT(a, b, c) \
  for (i = 0; i < 16; i++) \
    HASH_WHICH[i] = (i < a) ? 0 : ((i < a + b) ? 1 : 2);
  if (b2) /* 4 + 2 + 1 */
    WHICH_SPLIT (9, 5, 2);
  if (!b2 && b1 && b1 == b0 - 1) /* 2 + 1 */
    WHICH_SPLIT (11, 5, 0);
  if (!b2 && b1 && b1 == b0 - 2) /* 4 + 1 */
    WHICH_SPLIT (13, 3, 0);
  if (!b2 && !b1)
    WHICH_SPLIT (16, 0, 0);
  for (i = 0; i < 3; i++)
    if (MASK_ARR[i])
      MEMALIGN (TABLE_ARR[i], 64, (MASK_ARR[i] + 4) * sizeof (typeHash));
  HashClear ();
  return MB;
#endif
}

/******************************** HASH ********************************/

void HashLowerALLNew (uint64 Z, int move, int depth, int Value)
{
  int DEPTH, i;
  TRANS_DECLARE ();
  int max = 0, w = 0;
  YUSUF_DECLARE ();
  move &= 0x7fff;
  Trans = HASH_POINTER (Z);
  for (i = 0; i < 4; i++, Trans++)
    {
      HYATT_HASH (Trans, trans);
      if ((trans->hash ^ (Z >> 32)) == 0
	  && (!trans->DepthLower || IsALL (trans))
	  && trans->DepthLower <= depth)
	{
	  trans->DepthLower = depth;
	  trans->move = move;
	  trans->ValueLower = Value;
	  trans->age = AGE;
	  trans->flags |= FLAG_LOWER | FLAG_ALL;
	  TRACE (TRACE_HASH, HASH_WRITE ("LA", trans));
	  HYATT_HASH_WRITE (trans, Trans);
	  return;
	}
      DEPTH = MAX (trans->DepthLower, trans->DepthUpper);
      if (AGE_DEPTH_MIX (trans->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (trans->age, DEPTH);
	  w = i;
	}
    }
  Trans = HASH_POINTER (Z) + w;
  trans->hash = (Z >> 32);
  trans->DepthUpper = 0;
  trans->ValueUpper = 0;
  trans->DepthLower = depth;
  trans->move = move;
  trans->ValueLower = Value;
  trans->age = AGE;
  trans->flags = FLAG_LOWER | FLAG_ALL;
  TRACE (TRACE_HASH, HASH_OVERWRITE ("LA", trans));
  HYATT_HASH_WRITE (trans, Trans);
  return;
}

void HashUpperCUTNew (uint64 Z, int depth, int Value)
{
  int DEPTH, i;
  TRANS_DECLARE ();
  int max = 0, w = 0;
  YUSUF_DECLARE ();
  Trans = HASH_POINTER (Z);
  for (i = 0; i < 4; i++, Trans++)
    {
      HYATT_HASH (Trans, trans);
      if (!(trans->hash ^ (Z >> 32))
	  && (!trans->DepthUpper || IsCUT (trans))
	  && trans->DepthUpper <= depth)
	{
	  trans->DepthUpper = depth;
	  trans->ValueUpper = Value;
	  trans->age = AGE;
	  trans->flags |= FLAG_UPPER | FLAG_CUT;
          TRACE (TRACE_HASH, HASH_WRITE ("UC", trans));
	  HYATT_HASH_WRITE (trans, Trans);
	  return;
	}
      DEPTH = MAX (trans->DepthLower, trans->DepthUpper);
      if (AGE_DEPTH_MIX (trans->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (trans->age, DEPTH);
	  w = i;
	}
    }
  Trans = HASH_POINTER (Z) + w;
  trans->hash = (Z >> 32);
  trans->DepthLower = 0;
  trans->move = 0;
  trans->ValueLower = 0;
  trans->DepthUpper = depth;
  trans->ValueUpper = Value;
  trans->age = AGE;
  trans->flags = FLAG_UPPER | FLAG_CUT;
  TRACE (TRACE_HASH, HASH_OVERWRITE ("UC", trans));
  HYATT_HASH_WRITE (trans, Trans);
  return;
}

void HashLowerNew (uint64 Z, int move, int depth, int Value, int FLAGS)
{
  int DEPTH, i;
  TRANS_DECLARE ();
  int max = 0, w = 0;
  YUSUF_DECLARE ();
  move &= 0x7fff;
  Trans = HASH_POINTER (Z);
  for (i = 0; i < 4; i++, Trans++)
    {
      HYATT_HASH (Trans, trans);
      if (!(trans->hash ^ (Z >> 32)) && !IsExact (trans)
	  && trans->DepthLower <= depth)
	{
	  trans->DepthLower = depth;
	  trans->move = move;
	  trans->ValueLower = Value;
	  trans->age = AGE;
	  trans->flags &= ~(FLAG_ALL | LOWER_FROM_PV);
	  trans->flags |= FLAG_LOWER | FLAGS;
          TRACE (TRACE_HASH, HASH_WRITE ("L", trans));
	  HYATT_HASH_WRITE (trans, Trans);
	  return;
	}
      DEPTH = MAX (trans->DepthLower, trans->DepthUpper);
      if (AGE_DEPTH_MIX (trans->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (trans->age, DEPTH);
	  w = i;
	}
    }
  Trans = HASH_POINTER (Z) + w;
  trans->hash = (Z >> 32);
  trans->DepthUpper = 0;
  trans->ValueUpper = 0;
  trans->DepthLower = depth;
  trans->move = move;
  trans->ValueLower = Value;
  trans->age = AGE;
  trans->flags = FLAG_LOWER | FLAGS;
  TRACE (TRACE_HASH, HASH_OVERWRITE ("L", trans));
  HYATT_HASH_WRITE (trans, Trans);
  return;
}

void HashUpperNew (uint64 Z, int depth, int Value, int FLAGS)
{
  int DEPTH, i;
  TRANS_DECLARE ();
  int max = 0, w = 0;
  YUSUF_DECLARE ();
  Trans = HASH_POINTER (Z);
  for (i = 0; i < 4; i++, Trans++)
    {
      HYATT_HASH (Trans, trans);
      if (!(trans->hash ^ (Z >> 32)) && !IsExact (trans)
	  && trans->DepthUpper <= depth)
	{
	  trans->DepthUpper = depth;
	  trans->ValueUpper = Value;
	  trans->age = AGE;
	  trans->flags &= ~(FLAG_CUT | UPPER_FROM_PV);
	  trans->flags |= FLAG_UPPER | FLAGS;
          TRACE (TRACE_HASH, HASH_WRITE ("U", trans));
	  HYATT_HASH_WRITE (trans, Trans);
	  return;
	}
      DEPTH = MAX (trans->DepthLower, trans->DepthUpper);
      if (AGE_DEPTH_MIX (trans->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (trans->age, DEPTH);
	  w = i;
	}
    }
  Trans = HASH_POINTER (Z) + w;
  trans->hash = (Z >> 32);
  trans->DepthLower = 0;
  trans->move = 0;
  trans->ValueLower = 0;
  trans->DepthUpper = depth;
  trans->ValueUpper = Value;
  trans->age = AGE;
  trans->flags = FLAG_UPPER | FLAGS;
  TRACE (TRACE_HASH, HASH_OVERWRITE ("U", trans));
  HYATT_HASH_WRITE (trans, Trans);
  return;
}

static void pv_zobrist (uint64 Z, int move, int depth, int Value)
{
  int i, k = Z & PVHashMask;
  TRANS_PV_DECLARE ();
  int w = 0, max = 0;
  YUSUF_PV_DECLARE ();
  for (i = 0; i < 4; i++)
    {
      Trans_pv = PVHashTable + (k + i);
      HYATT_HASH (Trans_pv, trans_pv);
      if (trans_pv->hash == Z)
	{
	  trans_pv->depth = depth; /* esta seguro que depth es mejor ? */
	  trans_pv->Value = Value;
	  trans_pv->move = move;
	  trans_pv->age = AGE;
	  HYATT_HASH_WRITE (trans_pv, Trans_pv);
	  return;
	}
      if (AGE_DEPTH_MIX (trans_pv->age, trans_pv->depth) > max)
	{
	  max = AGE_DEPTH_MIX (trans_pv->age, trans_pv->depth);
	  w = i;
	}
    }
  Trans_pv = PVHashTable + (k + w);
  trans_pv->hash = Z;
  trans_pv->depth = depth;
  trans_pv->move = move;
  trans_pv->Value = Value;
  trans_pv->age = AGE;
  HYATT_HASH_WRITE (trans_pv, Trans_pv);
}

void HashExactNew (uint64 Z, int move, int depth, int Value, int FLAGS)
{
  int DEPTH, i, j;
  TRANS_DECLARE ();
  int max = 0, w = 0;
  YUSUF_DECLARE ();
  move &= 0x7fff;
  pv_zobrist (Z, move, depth, Value);
  Trans = HASH_POINTER (Z);
  for (i = 0; i < 4; i++, Trans++)
    {
      HYATT_HASH (Trans, trans);
      if ((trans->hash ^ (Z >> 32)) == 0
	  && MAX (trans->DepthUpper, trans->DepthLower) <= depth)
	{
	  trans->DepthUpper = trans->DepthLower = depth;
	  trans->move = move;
	  trans->ValueUpper = trans->ValueLower = Value;
	  trans->age = AGE;
	  trans->flags = FLAGS;
          TRACE (TRACE_HASH, HASH_WRITE ("E", trans));
	  HYATT_HASH_WRITE (trans, Trans);
	  for (j = i + 1; j < 4; j++)
	    {
	      Trans++;
	      HYATT_HASH (Trans, trans);
	      if ((trans->hash ^ (Z >> 32)) == 0
		  && MAX (trans->DepthUpper, trans->DepthLower) <= depth)
		{
		  memset (trans, 0, 16);
		  trans->age = AGE ^ (MAX_AGE / 2);
		  TRACE (TRACE_HASH, HASH_ZERO (trans));
		  HYATT_HASH_WRITE (trans, Trans);
		}
	    }
	  return;
	}
      DEPTH = MAX (trans->DepthLower, trans->DepthUpper);
      if (AGE_DEPTH_MIX (trans->age, DEPTH) > max)
	{
	  max = AGE_DEPTH_MIX (trans->age, DEPTH);
	  w = i;
	}
    }
  Trans = HASH_POINTER (Z) + w;
  trans->hash = (Z >> 32);
  trans->DepthUpper = trans->DepthLower = depth;
  trans->move = move;
  trans->ValueUpper = trans->ValueLower = Value;
  trans->age = AGE;
  trans->flags = FLAGS;
  TRACE (TRACE_HASH, HASH_OVERWRITE ("E", trans));
  HYATT_HASH_WRITE (trans, Trans);
  return;
}
