#include "RobboLito.h"
#include <string.h>
#include "control.h"

#if 0 /* UCI? */
int GetNumCPU() /*Get number of CPUs */
{
#ifdef WINDOWS /*Windows*/
  SYSTEM_INFO si;
  GetSystemInfo( &si );
  CPUS_SIZE = si.dwNumberOfProcessors;
#else /*LINUX */
#include <unistd.h>
  CPUS_SIZE = sysconf( _SC_NPROCESSORS_CONF );
#endif
  /*
#ifdef APPLE
#include <sys/sysctl.h>
CPUS_SIZE = 1;
size_t size=sizeof(CPUS_SIZE) ;
sysctlbyname("hw.ncpu",&CPUS_SIZE,&size,NULL,0);
#endif
  */
  return CPUS_SIZE;
} /*End Get number of CPUs*/
#endif

#if defined(UTILITIES)
#define SWAP_TWO(x, y) { (x) ^= (y); (y) ^= (x); (x) ^= (y); }
#define SHUFFLE_TWO(x) { (x) = ((((x) & 0x3) << 2) | (((x) & 0xc) >> 2)); }
static void InvertBoard (typePOS* POSITION)
{
  int f, r, sq;
  for (f = FA; f <= FH; f++)
    for (r = R1; r <= R4; r++)
      SWAP_TWO (POSITION->sq[8 * r + f], POSITION->sq[8 * (7 - r) + f]);
  for (sq = A1; sq <= H8; sq++)
    {
      if (POSITION->sq[sq])
	POSITION->sq[sq] ^= 8;
      if ((POSITION->sq[sq] & 0x6) == 0x4)
	POSITION->sq[sq] ^= 1; /* HACK */
    }
  POSITION->wtm ^= 1;
  if (POSITION->DYN->ep)
    POSITION->DYN->ep ^= 6 << 3; /* HACK */
  SHUFFLE_TWO (POSITION->DYN->oo);
  InitBitboards (POSITION);
}

static void EvalTest (typePOS* POSITION)
{
  int v;
  EvalExplanation (POSITION, -32767, 32767, 0);
  v = POSITION->DYN->Value;
  Eval (POSITION, -32767, 32767, 0);
  if (v != POSITION->DYN->Value)
    {
      printf ("ERROR with EvalExplanation compared to EvalEvaluation %d %d\n",
	      v, POSITION->DYN->Value);
    }
  InvertBoard (POSITION);
  Eval (POSITION, -32767, 32767, 0);
  if (v != POSITION->DYN->Value)
    {
      printf ("ERROR in InvertBoard Eval compared to Invert %d %d\n",
	      v, POSITION->DYN->Value);
      printf ("FEN is: %s\n", EmitFen (POSITION, STRING1));
    }
  InvertBoard (POSITION);
}
#endif

#if 0
#define READYOK() readyok ()
#else
#define READYOK() /* */
#endif

typedef enum { UCI_SPIN, UCI_CHECK, UCI_BUTTON, UCI_STRING } enum_UCI;
typedef struct
{char name[0x100]; int type; sint32 min,max,def;
 void* var; int (*action)(int);} UCI_type;
UCI_type UCI_OPTIONS[256]=
  {
    { "Hash", UCI_SPIN, 1, 65536, 32, NULL, &InitHash },
    { "PawnsHash", UCI_SPIN, 1, 1024, 2, NULL, &InitPawnHash },
    { "PVHash", UCI_SPIN, 1, 1024, 1, NULL, &InitPVHash },
    { "TitanicMode", UCI_CHECK, 0, 0, FALSE, &TITANIC_MODE, &ivan_init_smp },
    { "TitanicCPUsSizings", UCI_SPIN, 1, MAX_CPUS, 1, &CPUS_SIZE, &ivan_init_smp },
    { "Ponder", UCI_CHECK, 0, 0, FALSE, &UCI_PONDER, NULL },
    { "VerifyNullMove", UCI_CHECK, 0, 0, TRUE, &VERIFY_NULL, NULL },
    { "AlternativeTimeUsage", UCI_CHECK, 0, 0, FALSE, &USE_ALTERNATIVE_TIME, NULL },
    { "AllowInstantMoveFromHash", UCI_CHECK, 0, 0, TRUE, &ALLOW_INSTANT_MOVE, NULL },
    { "BufferTime", UCI_SPIN, 0, 10000, 0, &BUFFER_TIME, NULL},
    { "OutputDelay", UCI_SPIN, 0, 60000, 0, &OUTPUT_DELAY, NULL},
#ifdef CON_ROBBO_BUILD
    { "RobboTripleBaseDirectory", UCI_STRING, 0, 0, 0, ROBBO_TRIPLE_DIR, NULL},
    { "LoadRobboTripleBases", UCI_BUTTON, 0, 0, 0, NULL, &LoadRobboTripleBases},
    { "RobboTotalBaseDirectory", UCI_STRING, 0, 0, 0, ROBBO_TOTAL_DIR, NULL},
    { "RegisterRobboTotalBases", UCI_BUTTON, 0, 0, 0, NULL, &RegisterRobboTotalBases},
    { "RobboTotalBaseCacheSize", UCI_SPIN, 1, 1024, 1, NULL, &SetTotalBaseCache},
#endif
    { "MultiPV", UCI_SPIN, 1, 250, 1, &MULTI_PV, NULL},
    { "MultiCentiPawnPV", UCI_SPIN, 1, 65535, 65535, &MULTI_CENTI_PAWN_PV, NULL},
    { "RandomCount", UCI_SPIN, 0, 8, 0, &RANDOM_COUNT, NULL},
    { "RandomBits", UCI_SPIN, 1, 3, 1, &RANDOM_BITS, NULL},
#ifdef USER_SPLIT
    { "SplitAtCUT", UCI_CHECK, 0, 0, TRUE, &SPLIT_AT_CUT, NULL},
    { "SplitDepthCUT", UCI_SPIN, 16, 20, 16, &CUT_SPLIT_DEPTH, NULL},
    { "SplitDepthALL", UCI_SPIN, 12, 20, 14, &ALL_SPLIT_DEPTH, NULL},
    { "SplitDepthPV", UCI_SPIN, 10, 20, 14, &PV_SPLIT_DEPTH, NULL},
#endif
    { "WhiteBishopPair", UCI_SPIN, 1, 150, 45, &UCI_White_Bishops_Scale, &InitMaterialValue},
    { "WhitePawn", UCI_SPIN, 50, 200, 100, &UCI_White_Pawn_Scale, &InitMaterialValue},
    { "WhiteKnight", UCI_SPIN, 150, 500, 300, &UCI_White_Knight_Scale, &InitMaterialValue},
    { "WhiteLightBishop", UCI_SPIN, 150, 500, 310, &UCI_White_Light_Scale, &InitMaterialValue},
    { "WhiteDarkBishop", UCI_SPIN, 150, 500, 310, &UCI_White_Dark_Scale, &InitMaterialValue},
    { "WhiteRook", UCI_SPIN, 300, 700, 500, &UCI_White_Rook_Scale, &InitMaterialValue},
    { "WhiteQueen", UCI_SPIN, 500, 1500, 950, &UCI_White_Queen_Scale, &InitMaterialValue},
    { "BlackBishopPair", UCI_SPIN, 1, 150, 45, &UCI_Black_Bishops_Scale, &InitMaterialValue},
    { "BlackPawn", UCI_SPIN, 50, 200, 100, &UCI_Black_Pawn_Scale, &InitMaterialValue},
    { "BlackKnight", UCI_SPIN, 150, 500, 300, &UCI_Black_Knight_Scale, &InitMaterialValue},
    { "BlackLightBishop", UCI_SPIN, 150, 500, 310, &UCI_Black_Light_Scale, &InitMaterialValue},
    { "BlackDarkBishop", UCI_SPIN, 150, 500, 310, &UCI_Black_Dark_Scale, &InitMaterialValue},
    { "BlackRook", UCI_SPIN, 300, 700, 500, &UCI_Black_Rook_Scale, &InitMaterialValue},
    { "BlackQueen", UCI_SPIN, 500, 1500, 950, &UCI_Black_Queen_Scale, &InitMaterialValue},
    { "MaterialWeighting", UCI_SPIN, 512, 2048, 1024, &UCI_MATERIAL_WEIGHTING, &InitMaterialValue},
    { "KingSafetyWeighting", UCI_SPIN, 512, 2048, 1024, &UCI_KING_SAFETY_WEIGHTING, NULL},
    { "PawnsWeighting", UCI_SPIN, 512, 2048, 1024, &UCI_PAWNS_WEIGHTING, &PawnHashReset},
    { "AlwaysAnalyze", UCI_CHECK, 0, 0, FALSE, &ALWAYS_ANALYZE, NULL},
    { "TryPVinAnalysis", UCI_CHECK, 0, 0, TRUE, &TRY_PV_IN_ANALYSIS, NULL},
    { "SendCurrmove", UCI_CHECK, 0, 0, FALSE, &SEND_CURR_MOVE, NULL},
    { "DoHashFull", UCI_CHECK, 0, 0, TRUE, &DO_HASH_FULL, NULL},
#ifdef UTILITIES
    { "GetFEN", UCI_BUTTON, 0, 0, 0, NULL, &EmitFEN}, /* FIX ME */
    { "DebugTimeManagement", UCI_CHECK, 0, 0, FALSE, &DEBUG_TIME_MANAGEMENT, NULL},
#endif
    { "TimeEasyFactor", UCI_SPIN, 10, 100, 15, &EASY_FACTOR, NULL},
    { "TimeEasyFactorPonder", UCI_SPIN, 10, 100, 33, &EASY_FACTOR_PONDER, NULL},
    { "TimeBattleFactor", UCI_SPIN, 10, 500, 100, &BATTLE_FACTOR, NULL},
    { "TimeOrdinaryFactor", UCI_SPIN, 10, 500, 75, &ORDINARY_FACTOR, NULL},
    { "TimeAbsolutePercent", UCI_SPIN, 10, 100, 25, &ABSOLUTE_PERCENT, NULL},
    { "TimeDesiredMillis", UCI_SPIN, 10, 1000, 40, &DESIRED_MILLIS, NULL},
    { "TimeBookExitMoves", UCI_SPIN, 0, 20, 0, &BOOK_EXIT_MOVES, NULL},
    { "ExtraExtendInCheck", UCI_CHECK, 0, 0, FALSE, &EXTEND_IN_CHECK, NULL},
    { "", -1, 0, 0, FALSE, NULL, NULL}
  };

static void uciNew ()
{
  UCI_type* uci;
  char string[256];
  char* str;
  char TYPE[4][16] = {"spin", "check", "button", "string" };
  for (uci = UCI_OPTIONS; uci->type != -1; uci++) /* HACK */
    {
      str = string;
      sprintf (str, "option name %s type %s", uci->name, TYPE[uci->type]);
      str += strlen (str);
      if (uci->type == UCI_BUTTON)
	continue;
      if (uci->type == UCI_CHECK)
	sprintf (str, " default %s", uci->def ? "true" : "false");
      if (uci->type == UCI_STRING)
	sprintf (str, " default NULL");
      if (uci->type == UCI_SPIN)
	sprintf (str, " min %d max %d default %d", uci->min, uci->max, uci->def);
      str += strlen (str);
      sprintf (str, "\n");
      SEND ("%s", string);
    }
  SEND ("uciok\n");
}

static void SetOptionNew (char* string)
 {
   UCI_type* uci;
   char* v;
   int Value, r;
   for (uci = UCI_OPTIONS; uci->type != -1; uci++) /* HACK */
     {
       if (!memcmp(uci->name, string, strlen (uci->name)))
	 {
	   if (uci->type == UCI_BUTTON)
	     {
	       Value = uci->action (0);
	       SEND ("info string Button %s pushed, action worked %s\n",
		     uci->name, Value ? "true" : "false");
	       return;
	     }
	   v = strstr (string, "value");
	   if (!v)
	     return;
	   if (uci->type == UCI_STRING)
	     {
	       strcpy ((char*) (uci->var), v + 6);
	       SEND ("info string Optional %s %s\n", uci->name, (char*) (uci->var));
	       READYOK ();
	     }
	   if (uci->type == UCI_CHECK)
	     {
	       if (!strcmp (v + 6, "false"))
		 {
		   *((boolean*) (uci->var)) = FALSE;
		   if (uci->action)
		     uci->action (0);
		 }
	       if (!strcmp (v + 6, "true"))
		 {
		   *((boolean*) (uci->var)) = TRUE;
		   if (uci->action)
		     uci->action (0);
		 }
	       SEND ("info string Optional %s %s\n",
		     uci->name, (*((boolean*) (uci->var))) ? "true" : "false");
	       READYOK ();
	     }
	   if (uci->type == UCI_SPIN)
	     {
	       Value = atoi (v + 6);
	       if (Value < uci->min)
		 Value = uci->min;
	       if (Value > uci->max)
		 Value = uci->max;
	       if (uci->var)
		 *((sint32*) (uci->var)) = Value;
	       if (uci->action)
		 {	    
		   r = uci->action (Value);
		   if (r)
		     Value = r;
		 }
	       SEND ("info string Optional %s %d\n", uci->name, Value);
	     }
	   return;
	 }
     }
   SEND ("info string Optional unknown: %s\n", string);
 }

static void SetOption (typePOS* POSITION, char* string)
{
  char *v;
  int Value;
  if (!memcmp (string, "Hash", 4) || !memcmp (string, "TranspositionalTabularSizing", 28))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  Value = InitHash (Value);
	  SEND ("Optional TranspositionalTabularSizing %d\n", Value);
	  READYOK ();
	}
    }
  if (!memcmp (string, "PawnsHash", 9) || !memcmp (string, "PawnsZobrist", 12))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  Value = InitPawnHash (Value);
	  SEND ("Optional PawnsZobrist %d\n", Value);
	  READYOK ();
	}
    }
  if (!memcmp (string, "PVHash", 6))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  Value = InitPVHash (Value);
	  SEND ("Optional PVHash %d\n", Value);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TitanicCPUsSizings", 18))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value <= 0)
	    Value = 1;
	  if (Value > 64)
	    Value = 64;
	  if (Value > MAX_CPUS)
	    Value = MAX_CPUS;
	  CPUS_SIZE = Value;
	  if (TITANIC_MODE)
	    ivan_init_smp ();
	  SEND ("Optional TitanicCPUsSizings (Capitalist!) %d\n", Value);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TitanicMode", 11))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    TITANIC_MODE = FALSE;
	  if (!strcmp (v + 6, "true"))
	    {
	      TITANIC_MODE = TRUE;
	      ivan_init_smp ();
	    }
	  READYOK ();
	}
    }
  if (!memcmp (string, "BufferTime", 10))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 0)
	    Value = 0;
	  if (Value > 10000)
	    Value = 10000;
	  BUFFER_TIME = Value;
	  SEND ("Optional BufferTime %d\n", BUFFER_TIME);
	  READYOK ();
	}
    }
  if (!memcmp (string, "OutputDelay", 11))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 0)
	    Value = 0;
	  if (Value > 60000)
	    Value = 60000;
	  OUTPUT_DELAY = Value;
	  SEND ("Optional OutputDelay %d\n", OUTPUT_DELAY);
	  READYOK ();
	}
    }
  if (!memcmp (string, "AlternativeTimeUsage", 20))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    USE_ALTERNATIVE_TIME = FALSE;
	  if (!strcmp (v + 6, "true"))
	    USE_ALTERNATIVE_TIME = TRUE;
	  SEND ("Optional AlternativeTimeUsage %s\n",
		USE_ALTERNATIVE_TIME ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "AllowInstantMoveFromHash", 24))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    ALLOW_INSTANT_MOVE = FALSE;
	  if (!strcmp (v + 6, "true"))
	    ALLOW_INSTANT_MOVE = TRUE;
	  SEND ("Optional AllowInstantMoveFromHash %s\n",
		ALLOW_INSTANT_MOVE ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "Ponder", 6))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    UCI_PONDER = FALSE;
	  if (!strcmp (v + 6, "true"))
	    UCI_PONDER = TRUE;
	  SEND ("Optional Ponder %s\n", UCI_PONDER ? "true" : "false");
	  READYOK ();
	}
    }
  /********************************/  /********************************/
  if (!memcmp (string, "UCI_White_Bishop_Pair_Scale", 27))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 0)
	    Value = 0;
	  if (Value > 150)
	    Value = 150;
	  UCI_White_Bishops_Scale = Value;
	  SEND ("Optional UCI_White_Bishop_Pair_Scale %d\n", UCI_White_Bishops_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_White_Pawn_Scale", 20))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 50)
	    Value = 50;
	  if (Value > 200)
	    Value = 200;
	  UCI_White_Pawn_Scale = Value;
	  SEND ("Optional UCI_White_Pawn_Scale %d\n", UCI_White_Pawn_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_White_Knight_Scale", 22))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 150)
	    Value = 150;
	  if (Value > 500)
	    Value = 500;
	  UCI_White_Knight_Scale = Value;
	  SEND ("Optional UCI_White_Knight_Scale %d\n", UCI_White_Knight_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_White_Light_Bishop_Scale", 28))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 150)
	    Value = 150;
	  if (Value > 500)
	    Value = 500;
	  UCI_White_Light_Scale = Value;
	  SEND ("Optional UCI_White_Light_Bishop_Scale %d\n", UCI_White_Light_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_White_Dark_Bishop_Scale", 27))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 150)
	    Value = 150;
	  if (Value > 500)
	    Value = 500;
	  UCI_White_Dark_Scale = Value;
	  SEND ("Optional UCI_White_Dark_Bishop_Scale %d\n", UCI_White_Dark_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_White_Rook_Scale", 20))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 300)
	    Value = 300;
	  if (Value > 700)
	    Value = 700;
	  UCI_White_Rook_Scale = Value;
	  SEND ("Optional UCI_White_Rook_Scale %d\n", UCI_White_Rook_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_White_Queen_Scale", 21))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 500)
	    Value = 500;
	  if (Value > 1500)
	    Value = 1500;
	  UCI_White_Queen_Scale = Value;
	  SEND ("Optional UCI_White_Queen_Scale %d\n", UCI_White_Queen_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  /********************************/  /********************************/
  if (!memcmp (string, "UCI_Black_Bishop_Pair_Scale", 27))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 0)
	    Value = 0;
	  if (Value > 150)
	    Value = 150;
	  UCI_Black_Bishops_Scale = Value;
	  SEND ("Optional UCI_Black_Bishop_Pair_Scale %d\n", UCI_Black_Bishops_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_Black_Pawn_Scale", 20))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 50)
	    Value = 50;
	  if (Value > 200)
	    Value = 200;
	  UCI_Black_Pawn_Scale = Value;
	  SEND ("Optional UCI_Black_Pawn_Scale %d\n", UCI_Black_Pawn_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_Black_Knight_Scale", 22))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 150)
	    Value = 150;
	  if (Value > 500)
	    Value = 500;
	  UCI_Black_Knight_Scale = Value;
	  SEND ("Optional UCI_Black_Knight_Scale %d\n", UCI_Black_Knight_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_Black_Light_Bishop_Scale", 28))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 150)
	    Value = 150;
	  if (Value > 500)
	    Value = 500;
	  UCI_Black_Light_Scale = Value;
	  SEND ("Optional UCI_Black_Light_Bishop_Scale %d\n", UCI_Black_Light_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_Black_Dark_Bishop_Scale", 27))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 150)
	    Value = 150;
	  if (Value > 500)
	    Value = 500;
	  UCI_Black_Dark_Scale = Value;
	  SEND ("Optional UCI_Black_Dark_Bishop_Scale %d\n", UCI_Black_Dark_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_Black_Rook_Scale", 20))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 300)
	    Value = 300;
	  if (Value > 700)
	    Value = 700;
	  UCI_Black_Rook_Scale = Value;
	  SEND ("Optional UCI_Black_Rook_Scale %d\n", UCI_Black_Rook_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "UCI_Black_Queen_Scale", 21))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 500)
	    Value = 500;
	  if (Value > 1500)
	    Value = 1500;
	  UCI_Black_Queen_Scale = Value;
	  SEND ("Optional UCI_Black_Queen_Scale %d\n", UCI_Black_Queen_Scale);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "MaterialWeighting", 17))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 512)
	    Value = 512;
	  if (Value > 2048)
	    Value = 2048;
	  UCI_MATERIAL_WEIGHTING = Value;
	  SEND ("Optional MaterialWeighting %d\n", UCI_MATERIAL_WEIGHTING);
	  InitMaterialValue ();
	  EvalHashClear ();
	  READYOK ();
	}
    }
  if (!memcmp (string, "KingSafetyWeighting", 19))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 512)
	    Value = 512;
	  if (Value > 2048)
	    Value = 2048;
	  UCI_KING_SAFETY_WEIGHTING = Value;
	  SEND ("Optional KingSafetyWeighting %d\n", UCI_KING_SAFETY_WEIGHTING);
	  READYOK ();
	}
    }
  if (!memcmp (string, "PawnsWeighting", 14))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value < 512)
	    Value = 512;
	  if (Value > 2048)
	    Value = 2048;
	  UCI_PAWNS_WEIGHTING = Value;
	  SEND ("Optional PawnsWeighting %d\n", UCI_PAWNS_WEIGHTING);
	  PawnHashReset ();
	  READYOK ();
	}
    }
  /********************************/  /********************************/
  if (!memcmp (string, "VerifyNullMove", 14))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    VERIFY_NULL = FALSE;
	  if (!strcmp (v + 6, "true"))
	    VERIFY_NULL = TRUE;
	  SEND ("Optional VerifyNullMove %s\n", VERIFY_NULL ? "true" : "false");
	  READYOK ();
	}
    }
  /********************************/  /********************************/
#ifdef CON_ROBBO_BUILD
  if (!memcmp (string, "RobboTripleBaseDirectory", 24))
    {
      v = strstr (string, "value");
      if (v)
	{
	  strcpy (ROBBO_TRIPLE_DIR, v + 6);
	  SEND ("Optional RobboTripleBaseDirectory %s\n", ROBBO_TRIPLE_DIR);
	  READYOK ();
	}
    }
  if (!memcmp (string, "LoadRobboTripleBases", 20))
    {
      LoadRobboTripleBases ();
      READYOK ();
    }
  if (!memcmp (string, "RobboTotalBaseDirectory", 23))
    {
      v = strstr (string, "value");
      if (v)
	{
	  strcpy (ROBBO_TOTAL_DIR, v + 6);
	  SEND ("Optional RobboTotalBaseDirectory %s\n", ROBBO_TOTAL_DIR);
	  READYOK ();
	}
    }
  if (!memcmp (string, "RegisterRobboTotalBases", 23))
    {
      RegisterRobboTotalBases ();
      READYOK ();
    }

  if (!memcmp (string, "RobboTotalBaseCacheSize", 23))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  SetTotalBaseCache (Value);
	  READYOK ();
	}
    }

#endif
#ifndef MINIMAL
  if (!memcmp (string, "AlwaysAnalyze", 13))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    ALWAYS_ANALYZE = FALSE;
	  if (!strcmp (v + 6, "true"))
	    ALWAYS_ANALYZE = TRUE;
	  SEND ("Optional AlwaysAnalyze %s\n",
		ALWAYS_ANALYZE ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "TryPVinAnalysis", 15))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    TRY_PV_IN_ANALYSIS = FALSE;
	  if (!strcmp (v + 6, "true"))
	    TRY_PV_IN_ANALYSIS = TRUE;
	  SEND ("Optional TryPVinAnalysis %s\n",
		TRY_PV_IN_ANALYSIS ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "SendCurrmove", 12))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    SEND_CURR_MOVE = FALSE;
	  if (!strcmp (v + 6, "true"))
	    SEND_CURR_MOVE = TRUE;
	  SEND ("Optional SendCurrmove %s\n",
		SEND_CURR_MOVE ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "DoHashFull", 10))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    DO_HASH_FULL = FALSE;
	  if (!strcmp (v + 6, "true"))
	    DO_HASH_FULL = TRUE;
	  SEND ("Optional DoHashFull %s\n",
		DO_HASH_FULL ? "true" : "false");
	  READYOK ();
	}
    }
#ifdef UTILITIES
  if (!memcmp (string, "DebugTimeManagement", 19))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    DEBUG_TIME_MANAGEMENT = FALSE;
	  if (!strcmp (v + 6, "true"))
	    DEBUG_TIME_MANAGEMENT = TRUE;
	  SEND ("Optional DebugTimeManagement %s\n",
		DEBUG_TIME_MANAGEMENT ? "true" : "false");
	  READYOK ();
	}
    }
#endif
  if (!memcmp (string, "GetFEN", 6))
    {
      EmitFen (POSITION, STRING1[POSITION->cpu]);
      SEND ("info string %s\n", STRING1[POSITION->cpu]);
      READYOK ();
    }
  if (!memcmp (string, "RandomCount", 11))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  RANDOM_COUNT = Value;
	  if (RANDOM_COUNT < 0)
	    RANDOM_COUNT = 0;
	  if (RANDOM_COUNT >= 8)
	    RANDOM_COUNT = 8;
	  SEND ("Optional RandomCount %d\n", RANDOM_COUNT);
	  READYOK ();
	}
    }
  if (!memcmp (string, "RandomBits", 10))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  RANDOM_BITS = Value;
	  if (RANDOM_BITS < 1)
	    RANDOM_BITS = 1;
	  if (RANDOM_BITS >= 3)
	    RANDOM_BITS = 3;
	  SEND ("Optional RandomBits %d\n", RANDOM_BITS);
	  READYOK ();
	}
    }

#ifdef USER_SPLIT
  if (!memcmp (string, "SplitAtCUT", 10))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    SPLIT_AT_CUT = FALSE;
	  if (!strcmp (v + 6, "true"))
	    SPLIT_AT_CUT = TRUE;
	  SEND ("Optional SplitAtCUT %s\n", SPLIT_AT_CUT ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "SplitDepthCUT", 13))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  CUT_SPLIT_DEPTH = Value;
	  if (CUT_SPLIT_DEPTH <= 16)
	    CUT_SPLIT_DEPTH = 16;
	  if (CUT_SPLIT_DEPTH >= 20)
	    CUT_SPLIT_DEPTH = 20;
	  SEND ("Optional SplitDepthCUT %d\n", CUT_SPLIT_DEPTH);
	  READYOK ();
	}
    }
  if (!memcmp (string, "SplitDepthALL", 13))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  ALL_SPLIT_DEPTH = Value;
	  if (ALL_SPLIT_DEPTH <= 12)
	    ALL_SPLIT_DEPTH = 12;
	  if (ALL_SPLIT_DEPTH >= 20)
	    ALL_SPLIT_DEPTH = 20;
	  SEND ("Optional SplitDepthALL %d\n", ALL_SPLIT_DEPTH);
	  READYOK ();
	}
    }
  if (!memcmp (string, "SplitDepthPV", 12))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  PV_SPLIT_DEPTH = Value;
	  if (PV_SPLIT_DEPTH <= 10)
	    PV_SPLIT_DEPTH = 10;
	  if (PV_SPLIT_DEPTH >= 20)
	    PV_SPLIT_DEPTH = 20;
	  SEND ("Optional SplitDepthPV %d\n", PV_SPLIT_DEPTH);
	  READYOK ();
	}
    }
#endif
  if (!memcmp (string, "MultiPV", 7))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  MULTI_PV = Value;
	  if (MULTI_PV <= 1)
	    MULTI_PV = 1;
	  if (MULTI_PV >= 250)
	    MULTI_PV = 250;
	  SEND ("Optional MultiPV %d\n", MULTI_PV);
	  READYOK ();
	}
    }
  if (!memcmp (string, "MultiCentiPawnPV", 16))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  MULTI_CENTI_PAWN_PV = Value;
	  if (MULTI_CENTI_PAWN_PV <= 1)
	    MULTI_CENTI_PAWN_PV = 1;
	  if (MULTI_CENTI_PAWN_PV >= 65535)
	    MULTI_CENTI_PAWN_PV = 65535;
	  SEND ("Optional MultiCentiPawnPV %d\n", MULTI_CENTI_PAWN_PV);
	  READYOK ();
	}
    }
#endif

  if (!memcmp (string, "TimeEasyFactor", 14))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  EASY_FACTOR = Value;
	  if (EASY_FACTOR <= 10)
	    EASY_FACTOR = 10;
	  if (EASY_FACTOR >= 500)
	    EASY_FACTOR = 500;
	  SEND ("Optional TimeEasyFactor %d\n", EASY_FACTOR);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TimeEasyFactorPonder", 20))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  EASY_FACTOR_PONDER = Value;
	  if (EASY_FACTOR_PONDER <= 10)
	    EASY_FACTOR_PONDER = 10;
	  if (EASY_FACTOR_PONDER >= 500)
	    EASY_FACTOR_PONDER = 500;
	  SEND ("Optional TimeEasyFactorPonder %d\n", EASY_FACTOR_PONDER);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TimeBattleFactor", 16))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  BATTLE_FACTOR = Value;
	  if (BATTLE_FACTOR <= 10)
	    BATTLE_FACTOR = 10;
	  if (BATTLE_FACTOR >= 500)
	    BATTLE_FACTOR = 500;
	  SEND ("Optional TimeBattleFactor %d\n", BATTLE_FACTOR);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TimeOrdinaryFactor", 18))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  ORDINARY_FACTOR = Value;
	  if (ORDINARY_FACTOR <= 10)
	    ORDINARY_FACTOR = 10;
	  if (ORDINARY_FACTOR >= 500)
	    ORDINARY_FACTOR = 500;
	  SEND ("Optional TimeOrdinaryFactor %d\n", ORDINARY_FACTOR);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TimeAbsolutePercent", 19))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  ABSOLUTE_PERCENT = Value;
	  if (ABSOLUTE_PERCENT <= 10)
	    ABSOLUTE_PERCENT = 10;
	  if (ABSOLUTE_PERCENT >= 100)
	    ABSOLUTE_PERCENT = 100;
	  SEND ("Optional TimeAbsolutePercent %d\n", ABSOLUTE_PERCENT);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TimeDesiredMillis", 17))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  DESIRED_MILLIS = Value;
	  if (DESIRED_MILLIS <= 10)
	    DESIRED_MILLIS = 10;
	  if (DESIRED_MILLIS >= 100)
	    DESIRED_MILLIS = 100;
	  SEND ("Optional TimeDesiredMillis %d\n", DESIRED_MILLIS);
	  READYOK ();
	}
    }
  if (!memcmp (string, "TimeBookExitMoves", 17))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  BOOK_EXIT_MOVES = Value;
	  if (BOOK_EXIT_MOVES <= 0)
	    BOOK_EXIT_MOVES = 0;
	  if (BOOK_EXIT_MOVES >= 20)
	    BOOK_EXIT_MOVES = 20;
	  SEND ("Optional TimeBookExitMoves %d\n", BOOK_EXIT_MOVES);
	  READYOK ();
	}
    }
  if (!memcmp (string, "ExtraExtendInCheck", 18))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    EXTEND_IN_CHECK = FALSE;
	  if (!strcmp (v + 6, "true"))
	    EXTEND_IN_CHECK = TRUE;
	  SEND ("Optional ExtraExtendInCheck %s\n",
		EXTEND_IN_CHECK ? "true" : "false");
	  READYOK ();
	}
    }

#ifdef BUILD_ZOG_MP_COMPILE
  if (!memcmp (string, "ZOG MP Usage", 12))
    {
      v = strstr (string, "value");
      if (v)
	{
	  if (!strcmp (v + 6, "false"))
	    ZOG_OFF ();
	  if (!strcmp (v + 6, "true"))
	    {
	      if (!ZOG_AKTIV)
		ZOG_ON ();
	    }
	  SEND ("Optional ZOG MP Usage %s\n", ZOG_AKTIV ? "true" : "false");
	  READYOK ();
	}
    }
  if (!memcmp (string, "ZOG MP Records", 14))
    {
      v = strstr (string, "value");
      if (v)
	{
	  Value = atoi (v + 6);
	  if (Value <= 1)
	    Value = 1;
	  if (Value >= 1024)
	    Value = 1024;
	  if (ZOG_AKTIV)
	    {
	      ZOG_OFF ();
	      ZOG_SIZE = Value << 10;
	      ZOG_ON ();
	    }
	  else
	    ZOG_SIZE = Value << 10;
	  SEND ("Optional ZOG MP Records %d\n", ZOG_SIZE >> 10);
	  READYOK ();
	}
    }
#endif
}

static void uci ()
{
  SEND ("id name %s version %s\n", EPONYM, VERSION);
  SEND ("id author Yakov Petrovich Golyadkin, ");
  SEND ("Igor Igorovich Igoronov, Roberto Pescatore, ");
  SEND ("Yusuf Ralf Weisskopf, ");
  SEND ("Ivan Skavinsky Skavar plus Decembrists (all)\n");
  SEND ("id copyright Yakov Petrovich Golyadkin, "
	"92th plus 93th year from Revolution, PUBLICDOMAIN (workers)\n");
  SEND ("id dedication To Vladimir Ilyich plus Zog 1 with Albania\n");
#ifndef OLD_UCI
  uciNew();
  return;
#endif
  SEND ("option name Hash type spin min 1 max 65536 default 32\n");
  SEND ("option name PawnsHash (binary mb) type spin min 1 max 1024 default 2\n");
  SEND ("option name PVHash (binary mb) type spin min 1 max 1024 default 1\n");
  SEND ("option name TitanicMode (Capitalist!) type check default false\n");
  SEND ("option name TitanicCPUsSizings (Capitalist!) type spin min 1 max %d default 1\n", MAX_CPUS);
  SEND ("option name Ponder type check default false\n");
  SEND ("option name VerifyNullMove type check default true\n");
  SEND ("option name AlternativeTimeUsage (beta) type check default false\n");
  SEND ("option name AllowInstantMoveFromHash (ponder off only) type check default true\n");
  SEND ("option name BufferTime (milliseconds) type spin min 0 max 10000 default 0\n");
  SEND ("option name OutputDelay (milliseconds) type spin min 0 max 60000 default 0\n");
#ifdef CON_ROBBO_BUILD
  SEND ("option name RobboTripleBaseDirectory type string default NULL\n");
  SEND ("option name LoadRobboTripleBases type button\n");
  SEND ("option name RobboTotalBaseDirectory type string default NULL\n");
  SEND ("option name RegisterRobboTotalBases type button\n");
  SEND ("option name RobboTotalBaseCacheSize type spin min 1 max 1024 default 1\n");
#endif
#ifdef UTILITIES
  SEND ("option name DebugTimeManagement type check default false\n");
#endif
#ifndef MINIMAL
  SEND ("option name MultiPV type spin min 1 max 250 default 1\n");
  SEND ("option name MultiCentiPawnPV type spin min 1 max 65535 default 65535\n");
  SEND ("option name RandomCount type spin min 0 max 8 default 0\n");
  SEND ("option name RandomBits type spin min 1 max 3 default 1\n");
#ifdef USER_SPLIT
  SEND ("option name SplitAtCUT (half ply) type check default true\n");
  SEND ("option name SplitDepthCUT (half ply) type spin min 16 max 20 default 16\n");
  SEND ("option name SplitDepthALL (half ply) type spin min 12 max 20 default 14\n");
  SEND ("option name SplitDepthPV (half ply) type spin min 10 max 20 default 14\n");
#endif
  SEND ("option name UCI_White_Bishop_Pair_Scale (cp) type spin min 0 max 150 default 45\n");
  SEND ("option name UCI_White_Pawn_Scale (cp) type spin min 50 max 200 default 100\n");
  SEND ("option name UCI_White_Knight_Scale (cp) type spin min 150 max 500 default 300\n");
  SEND ("option name UCI_White_Light_Bishop_Scale (cp) type spin min 150 max 500 default 310\n");
  SEND ("option name UCI_White_Dark_Bishop_Scale (cp) type spin min 150 max 500 default 310\n");
  SEND ("option name UCI_White_Rook_Scale (cp) type spin min 300 max 700 default 500\n");
  SEND ("option name UCI_White_Queen_Scale (cp) type spin min 500 max 1500 default 950\n");
  SEND ("option name UCI_Black_Bishop_Pair_Scale (cp) type spin min 0 max 150 default 45\n");
  SEND ("option name UCI_Black_Pawn_Scale (cp) type spin min 50 max 200 default 100\n");
  SEND ("option name UCI_Black_Knight_Scale (cp) type spin min 150 max 500 default 300\n");
  SEND ("option name UCI_Black_Light_Bishop_Scale (cp) type spin min 150 max 500 default 310\n");
  SEND ("option name UCI_Black_Dark_Bishop_Scale (cp) type spin min 150 max 500 default 310\n");
  SEND ("option name UCI_Black_Rook_Scale (cp) type spin min 300 max 700 default 500\n");
  SEND ("option name UCI_Black_Queen_Scale (cp) type spin min 500 max 1500 default 950\n");
  SEND ("option name MaterialWeighting type spin min 512 max 2048 default 1024\n");
  SEND ("option name KingSafetyWeighting type spin min 512 max 2048 default 1024\n");
  SEND ("option name PawnsWeighting type spin min 512 max 2048 default 1024\n");
  SEND ("option name AlwaysAnalyze type check default false\n");
  SEND ("option name TryPVinAnalysis type check default true\n");
  SEND ("option name SendCurrmove type check default false\n");
  SEND ("option name DoHashFull type check default true\n");
  SEND ("option name GetFEN type button\n");
#endif
  SEND ("option name TimeEasyFactor type spin min 10 max 500 default 15\n");
  SEND ("option name TimeEasyFactorPonder type spin min 10 max 500 default 33\n");
  SEND ("option name TimeBattleFactor type spin min 10 max 500 default 100\n");
  SEND ("option name TimeOrdinaryFactor type spin min 10 max 500 default 75\n");
  SEND ("option name TimeAbsolutePercent type spin min 10 max 100 default 25\n");
  SEND ("option name TimeDesiredMillis type spin min 10 max 100 default 40\n");
  SEND ("option name TimeBookExitMoves type spin min 0 max 20 default 0\n");
  SEND ("option name ExtraExtendInCheck type check default false \n");
  
#ifdef BUILD_ZOG_MP_COMPILE
  SEND ("option name ZOG MP Usage type check default false\n");
  SEND ("option name ZOG MP Records (binary kilos) type spin min 1 max 1024 default 4\n");
#endif
  SEND ("uciok\n");
}

static void readyok ()
{
  SEND ("readyok\n");
}

static void quit ()
{
  exit (0);
}

static void ParseInput (typePOS* POSITION, char* I)
{
  UCI_NEW_GAME = FALSE;
  if (!strcmp (I, "quit"))
    quit ();
  if (!strcmp (I, "stop"))
    {
      if (SEARCH_IS_DONE)
	return;
      HaltSearch (0, 0); /* no return when JUMP_IS_SET */
      if (INFINITE_LOOP)
	return;
      STALL_MODE = TRUE;
      while (!SEARCH_IS_DONE)
	NANO_SLEEP (1000000);
      STALL_MODE = FALSE;
      SUPPRESS_INPUT = FALSE;
      return;
    }
  if (!strcmp (I, "isready"))
    {
      readyok ();
      return;
    }
  if (!strcmp (I, "ponderhit")) /* cuidado con ZOG? Personal Opinion: for the OK! */
    {
      ponderhit ();
      return;
    }
  if (!strcmp (I, "ucinewgame"))
    {
      if (SEARCH_IS_DONE)
	{
	  NewGame (POSITION, TRUE);
	  return;
	}
      UCI_NEW_GAME = TRUE;
      HaltSearch (0, 0); /* no return when JUMP_IS_SET */
      if (INFINITE_LOOP)
	return;
      STALL_MODE = TRUE;
      while (!SEARCH_IS_DONE)
	NANO_SLEEP (1000000);
      STALL_MODE = FALSE;
      SUPPRESS_INPUT = FALSE;
      return;
    }
  if (JUMP_IS_SET)
    return;
#if defined (BENCHMARK) || defined(UTILITIES)
  if (!strcmp (I, "benchmark"))
    BenchMark (POSITION, "go movetime 1000");
  else if (!memcmp (I, "benchmark", 9)) /* run benchmark */
    BenchMark (POSITION, I + 10);
  if (!strcmp (I, "makeundo"))
    MakeUndoSpeed (POSITION);
#endif
#if defined (UTILITIES)
  if (!memcmp (I, "triple-lookup", 13))
    {
      int r;
      boolean b;
      b = TRIPLE_VALUE (POSITION, &r);
      printf ("RobboTripleValue %d\n", !b ? -100 : r);
    }
  if (!memcmp (I, "total-lookup", 12))
    {
      int r;
      boolean b;
      boolean RobboTotalBaseScore (typePOS*, int*);
      b = RobboTotalBaseScore (POSITION, &r);
      printf ("RobboTotalValue %d\n", !b ? -100 : r);
    }
  if (!memcmp (I, "perft-check", 11))
    Perfd (POSITION, atoi (I + 12), atoi (I + 14)); /* HACK */
  if (!memcmp (I, "drawboard", 9))
    DrawBoard (POSITION);
  if (!strcmp (I, "eval"))
    EvalTest (POSITION);
  if (!memcmp (I, "perft ", 6))
    PERFD (POSITION, atoi (I + 6));
  if (!memcmp (I, "verify-triple", 13))
    {
      int p[4];
      char A[64], B[64], C[64], D[64];
      sscanf (I + 14, "%s %s %s %s", A, B, C, D);
      p[0] = atoi(A);
      p[1] = atoi(B);
      p[2] = atoi(C);
      p[3] = atoi(D);
      VerifyTripleBase (POSITION, p);
    }
#endif
  if (!memcmp (I, "go", 2))
    {
      InitSearch (POSITION, I);
      if (BOARD_IS_OK)
	Search (POSITION);
    }
  if (!memcmp (I, "position", 8))
    InitPosition (POSITION, I + 9);
  if (!memcmp (I, "setoption name", 14))
#ifdef OLD_UCI
    SetOption (POSITION, I + 15);
#else
    SetOptionNew (I + 15);
#endif
  if (!strcmp (I, "uci"))
    uci ();
  SUPPRESS_INPUT = FALSE; /* ??? */
}

#ifdef WINDOWS
#include <io.h>
#include <conio.h> /* ? */
void Input (typePOS* POSITION)
{
  int i, r = 0;
  if (SUPPRESS_INPUT)
    return;
  if (_kbhit()) /* does this work? */
    {
      INPUT_BUFFER[r] = (char) getche();
      while (INPUT_BUFFER[r++] != '\r') /* what is \r ?? */
	  INPUT_BUFFER[r] = (char) getche ();
    }
  else
    {
      _read (_fileno (stdin), INPUT_BUFFER, 1);
      while (INPUT_BUFFER[r++] != '\n') /* but now \n is correct? */
	_read (_fileno (stdin), INPUT_BUFFER + r, 1);
    }
  INPUT_BUFFER[r - 1] = 0; /* gobble \n, or maybe \r who knows? */
  INPUT_BUFFER[r] = 0; /* safety */
  ParseInput (POSITION, INPUT_BUFFER);
}
#else
#include <unistd.h> /* LINUX */
void Input (typePOS* POSITION)
{
  int i, r = 0;
  if (SUPPRESS_INPUT)
    return;
  read (STDIN_FILENO, INPUT_BUFFER, 1);
  while (INPUT_BUFFER[r++] != '\n') /* newline, is always? MAC OS X? */
    read (STDIN_FILENO, INPUT_BUFFER + r, 1);
  INPUT_BUFFER[r - 1] = 0; /* gobble \n */
  INPUT_BUFFER[r] = 0; /* safety */
  ParseInput (POSITION, INPUT_BUFFER);
}
#endif

