
#include "RobboLito.h"

int PawnHashReset ()
{
  memset (PawnHash, 0, PawnHashSize * sizeof (typePawnEval));
  return FALSE; /* HACK */
}

int InitPawnHash (int mb)
{
  if (mb > 1024)
    mb = 1024;
  PawnHashSize = ((1 << BSR (mb)) << 20) / sizeof (typePawnEval);
  mb = (PawnHashSize * sizeof (typePawnEval)) >> 20;
  if (PawnHash)
    ALIGNED_FREE (PawnHash);
  MEMALIGN (PawnHash, 64, PawnHashSize * sizeof (typePawnEval));
  PawnHashReset ();
  return (mb);
}

void ResetPositionalGain ()
{
  int p, m;
#ifdef MULTIPLE_POS_GAIN
  int cpu;
  for (cpu = 0; cpu < MAX_CPUS; cpu++)
#endif
    for (p = 0; p < 0x10; p++)
      for (m = 0; m < 010000; m++)
#ifdef MULTIPLE_POS_GAIN
	MAX_POSITIONAL_GAIN[cpu][p][m] = 0;
#else
        MAX_POSITIONAL_GAIN[p][m] = 0;
#endif
}

void ResetHistory ()
{
  int pi, sq;
#ifdef MULTIPLE_HISTORY
  int cpu;
  for (cpu = 0; cpu < MAX_CPUS; cpu++)
#endif
    for (pi = 0; pi < 0x10; pi++)
      for (sq = A1; sq <= H8; sq++)
#ifdef MULTIPLE_HISTORY
	HISTORY[cpu][pi][sq] = 0x800;
#else
	HISTORY[pi][sq] = 0x800;
#endif
}

static void Banner ()
{
  printf ("%s VERSION %s\n", EPONYM, VERSION);
#ifdef HAS_PREFETCH
  printf ("compiled with PREFETCH\n");
#endif
#ifdef HAS_POPCNT
  printf ("compiled with POPCNT\n");
#endif
#ifdef YUSUF_MULTICORE
  printf ("Capitalist multicore mode compiled\n");
#endif
#ifdef UTILITIES
  printf ("Mode with UTILITIES compiled\n");
#endif
#ifdef MINIMAL
  printf ("MINIMAL Mode compiled\n");
#endif
#ifdef CON_ROBBO_BUILD
  printf ("RobboBases enabled by compile\n");
#endif
#ifdef BUILD_ZOG_MP_COMPILE
  printf ("compiled with ZOG MP, sizeof %d+%d\n",
	  (int) (sizeof (type_zog_node)), (int) (2 * sizeof (typeDYNAMIC)));
#endif
}

int main ()
{
  Banner ();
  PawnHash = NULL;
  PawnHashSize = (1 << 16);
  memset ( (void*) ROOT_POSIZIONE0, 0, sizeof (typePOS));
  ROOT_POSIZIONE0->DYN_ROOT =
    malloc (MAXIMUM_PLY * sizeof (typeDYNAMIC));
  ROOT_POSIZIONE0->DYN = ROOT_POSIZIONE0->DYN_ROOT; 
  ROOT_POSIZIONE0->stop = FALSE;
  CPUS_SIZE = 1;
#ifdef YUSUF_MULTICORE
  rp_init ();
#endif
  SMP_HAS_AKTIV = FALSE;
  InitPawnHash (2);
  InitHash (32);
  InitPVHash (1);
  ResetHistory ();
  InitCaptureValues ();
  InitArrays ();
  InitPawns ();
  UCI_White_Bishops_Scale = 45;
  UCI_White_Pawn_Scale = 100;
  UCI_White_Knight_Scale = 300;
  UCI_White_Light_Scale = 310;
  UCI_White_Dark_Scale = 310;
  UCI_White_Rook_Scale = 500;
  UCI_White_Queen_Scale = 950;
  UCI_Black_Bishops_Scale = 45;
  UCI_Black_Pawn_Scale = 100;
  UCI_Black_Knight_Scale = 300;
  UCI_Black_Light_Scale = 310;
  UCI_Black_Dark_Scale = 310;
  UCI_Black_Rook_Scale = 500;
  UCI_Black_Queen_Scale = 950;
  UCI_MATERIAL_WEIGHTING = 1024;
  UCI_PAWNS_WEIGHTING = 1024;
  UCI_KING_SAFETY_WEIGHTING = 1024;
  InitMaterialValue ();
  InitStatic ();
  NewGame (ROOT_POSIZIONE0, TRUE);

#ifdef USER_SPLIT
  CUT_SPLIT_DEPTH = 16;
  ALL_SPLIT_DEPTH = 14;
  PV_SPLIT_DEPTH = 14;
  SPLIT_AT_CUT = TRUE;
#endif
  SEND_CURR_MOVE = FALSE;
  USE_ALTERNATIVE_TIME = FALSE;
  BUFFER_TIME = 0;
  OUTPUT_DELAY = 0;
  UCI_PONDER = FALSE;
  VERIFY_NULL = TRUE;
  ROBBO_LOAD = FALSE;
  SEARCH_ROBBO_BASES = TRUE;
  MULTI_PV = 1;
  SEND_HASH = TRUE;
  DO_HASH_FULL = TRUE;
  ALLOW_INSTANT_MOVE = TRUE;
  ZOG_AKTIV = FALSE;
  strcpy (ROBBO_TRIPLE_DIR, "");
  strcpy (ROBBO_TOTAL_DIR, "");
  SEARCH_IS_DONE = TRUE;
  INFINITE_LOOP = FALSE;
#ifndef MINIMAL
  MULTI_CENTI_PAWN_PV = 65535;
  RANDOM_COUNT = 0;
  RANDOM_BITS = 1;
  ALWAYS_ANALYZE = FALSE;
  TRY_PV_IN_ANALYSIS = TRUE;
#endif

#ifdef TRACE_COMPILE
  TRACE_QSEARCH = TRUE;
  TRACE_HASH = TRUE;
  TRACE_EVAL = TRUE;
  TRACE_LOW_DEPTH = TRUE;
  TRACE_CUT = TRUE;
  TRACE_ALL = TRUE;
  TRACE_EXCLUDE = TRUE;
  TRACE_PV = TRUE;
  TRACE_MAKE_UNMAKE = TRUE;
#endif

#ifdef CON_ROBBO_BUILD
  total_iniz ();
#endif
#ifdef BUILD_ZOG_MP_COMPILE
  ZOG_init ();
#endif
  SUPPRESS_INPUT = FALSE;
  INPUT_BUFFER = malloc (65536);
  input_ptr = INPUT_BUFFER;
  STALL_MODE = FALSE;
  SINCE_NEW_GAME = 0;
  EASY_FACTOR = 15;
  EASY_FACTOR_PONDER = 33;
  BATTLE_FACTOR = 100;
  ORDINARY_FACTOR = 75;
  ABSOLUTE_PERCENT = 25;
  DESIRED_MILLIS =  40;
  BOOK_EXIT_MOVES = 0;
  EXTEND_IN_CHECK = FALSE;
  while (1)
    Input (ROOT_POSIZIONE0);
  return 0;
}
